# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Literal
from typing import overload

import logging
from datetime import datetime
from datetime import timedelta
from datetime import UTC

from gi.repository import GLib
from gi.repository import Gtk
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.task import Task

from gajim.common import app
from gajim.common import ged
from gajim.common.const import ArchiveState
from gajim.common.events import ArchivingIntervalFinished
from gajim.common.events import RawMamMessageReceived
from gajim.common.i18n import _
from gajim.common.util.decorators import event_filter

from gajim.gtk.assistant import Assistant
from gajim.gtk.assistant import AssistantPage
from gajim.gtk.assistant import AssistantSuccessPage
from gajim.gtk.dropdown import GajimDropDown

log = logging.getLogger("gajim.gtk.history_sync")


class HistorySyncAssistant(Assistant):
    def __init__(self, account: str) -> None:
        Assistant.__init__(self, name="HistorySyncAssistant", width=600)

        self.account = account
        self._client = app.get_client(account)

        self._timedelta: timedelta | None = None
        self._now = datetime.now(UTC)
        self._query_id: str | None = None
        self._start: datetime = ArchiveState.ALL
        self._end: datetime | None = None

        mam_start = None
        archive = app.storage.archive.get_mam_archive_state(
            account, self._client.get_own_jid().new_as_bare()
        )

        if archive is not None and archive.from_stanza_ts is not None:
            mam_start = archive.from_stanza_ts

        if mam_start is None:
            self._current_start = self._now
        else:
            self._current_start = mam_start

        self.add_button("close", _("Close"))
        self.add_button("synchronize", _("Synchronize"), "suggested-action")
        self.set_button_visible_func(self._visible_func)

        self.add_pages(
            {
                "select": SelectTime(self._now, self._current_start),
                "progress": Progress(),
            }
        )

        self.add_default_page("success")
        success_page = self.get_page("success")
        success_page.set_title(_("Synchronize Chat History"))
        success_page.set_heading(_("Finished"))

        self._connect(self, "button-clicked", self._on_button_clicked)

        # pylint: disable=line-too-long
        self.register_events(
            [
                ("archiving-interval-finished", ged.GUI1, self._received_finished),
                ("raw-mam-message-received", ged.PRECORE, self._mam_message_received),
            ]
        )
        # pylint: enable=line-too-long

        if mam_start == ArchiveState.ALL:
            success_page.set_text(_("Gajim is fully synchronized with the archive."))
            self.show_page("success")

    def _cleanup(self) -> None:
        self.get_page("select").run_destroy()
        Assistant._cleanup(self)

    def get_active_query_id(self) -> str | None:
        return self._query_id

    @overload
    def get_page(self, name: Literal["select"]) -> SelectTime: ...

    @overload
    def get_page(self, name: Literal["progress"]) -> Progress: ...

    @overload
    def get_page(self, name: Literal["success"]) -> AssistantSuccessPage: ...

    def get_page(self, name: str) -> AssistantPage:
        return self._pages[name]

    @staticmethod
    def _visible_func(_assistant: Assistant, page_name: str) -> list[str]:
        if page_name == "select":
            return ["close", "synchronize"]

        if page_name == "progress":
            return ["close"]

        if page_name == "success":
            return ["close"]

        raise ValueError(f"page {page_name} unknown")

    def _on_button_clicked(self, _assistant: Assistant, button_name: str) -> None:
        if button_name == "synchronize":
            self._prepare_query()
            self.show_page("progress", Gtk.StackTransitionType.SLIDE_LEFT)
            return

        if button_name == "close":
            self.close()

    def _prepare_query(self) -> None:
        select_time_page = self.get_page("select")
        self._timedelta = select_time_page.get_timedelta()
        if self._timedelta is not None:
            self._start = self._now - self._timedelta
        self._end = self._current_start

        log.info("Get mam_start_date: %s", self._current_start)
        log.info("Now: %s", self._now)
        log.info("Start: %s", self._start)
        log.info("End: %s", self._end)

        # Archive count query
        self._client.get_module("MAM").make_query(
            self._client.get_own_jid().bare,
            start=self._start,
            end=self._end,
            max_=0,
            callback=self._received_count,
        )

    def _received_count(self, task: Task) -> None:
        try:
            result = task.finish()
        except (StanzaError, MalformedStanzaError):
            return

        if result.rsm.count is not None:
            self.get_page("progress").set_count(int(result.rsm.count))
        mam_module = self._client.get_module("MAM")
        assert self._start is not None
        assert self._end is not None
        self._query_id = mam_module.request_archive_interval(self._start, self._end)

    @event_filter(["account"])
    def _received_finished(self, event: ArchivingIntervalFinished) -> None:
        if event.query_id != self._query_id:
            return
        self._query_id = None
        log.info("Query finished")
        progress_page = self.get_page("progress")
        GLib.idle_add(progress_page.set_finished)
        received_count = progress_page.get_received_count()
        success_page = self.get_page("success")
        success_page.set_text(
            _("Finished synchronising chat history:\n%s messages downloaded")
            % received_count
        )
        self.show_page("success")

    @event_filter(["account"])
    def _mam_message_received(self, event: RawMamMessageReceived) -> None:
        if self._query_id != event.properties.mam.query_id:
            return

        log.debug("Received message")
        self.get_page("progress").set_fraction()


class SelectTime(AssistantPage):
    def __init__(self, now: datetime, current_start: datetime) -> None:
        AssistantPage.__init__(self)
        self.title = _("Synchronize Chat History")
        self.complete = False
        self._timedelta: timedelta | None = None

        heading = Gtk.Label(
            label=_("Synchronize Chat History"),
            max_width_chars=30,
            wrap=True,
            halign=Gtk.Align.CENTER,
            justify=Gtk.Justification.CENTER,
        )
        heading.add_css_class("title-1")

        label = Gtk.Label(
            label=_("How far back should the chat history be synchronized?"),
            halign=Gtk.Align.CENTER,
            wrap=True,
            max_width_chars=40,
        )

        data = {
            30: _("One Month"),
            90: _("Three Months"),
            365: _("One Year"),
            -1: _("Everything"),
        }

        for days in list(data.keys()):
            if days == -1:
                continue

            delta = timedelta(days)
            if now - delta > current_start:
                data.pop(days)

        self._dropdown: GajimDropDown[int] = GajimDropDown(data, fixed_width=20)
        self._dropdown.set_halign(Gtk.Align.CENTER)

        self.append(heading)
        self.append(label)
        self.append(self._dropdown)

        self.complete = True
        self.update_page_complete()

    def run_destroy(self) -> None:
        self._dropdown.run_destroy()

    def get_timedelta(self) -> timedelta | None:
        key = self._dropdown.get_selected_key()
        if key is None:
            raise ValueError("No value selected")

        if key == -1:
            return None

        return timedelta(key)

    def get_default_button(self) -> str:
        return "synchronize"


class Progress(AssistantPage):
    def __init__(self) -> None:
        AssistantPage.__init__(self)
        self.title = _("Synchronizing Chat History…")

        self._count = 0
        self._received = 0

        image = Gtk.Image.new_from_icon_name("lucide-file-down-symbolic")
        image.set_pixel_size(64)

        self._progress_bar = Gtk.ProgressBar()
        self._progress_bar.set_show_text(True)
        self._progress_bar.set_text(_("Connecting..."))
        self._progress_bar.set_pulse_step(0.1)
        self._progress_bar.set_vexpand(True)
        self._progress_bar.set_valign(Gtk.Align.CENTER)

        self.append(image)
        self.append(self._progress_bar)

    def set_fraction(self) -> None:
        self._received += 1
        if self._count:
            self._progress_bar.set_fraction(self._received / self._count)
            self._progress_bar.set_text(
                _("%(received)s of %(max)s")
                % {"received": self._received, "max": self._count}
            )
        else:
            self._progress_bar.pulse()
            self._progress_bar.set_text(_("Downloaded %s messages") % self._received)

    def set_count(self, count: int) -> None:
        self._count = count

    def set_finished(self) -> None:
        self._progress_bar.set_fraction(1)

    def get_received_count(self) -> int:
        return self._received



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import cast
from typing import Literal
from typing import TYPE_CHECKING
from typing import TypedDict

import logging

from nbxmpp.structs import StanzaHandler

from gajim.common import app
from gajim.common.modules.adhoc_commands import AdHocCommands
from gajim.common.modules.annotations import Annotations
from gajim.common.modules.base import BaseModule
from gajim.common.modules.bits_of_binary import BitsOfBinary
from gajim.common.modules.blocking import Blocking
from gajim.common.modules.bookmarks import Bookmarks
from gajim.common.modules.bytestream import Bytestream
from gajim.common.modules.caps import Caps
from gajim.common.modules.carbons import Carbons
from gajim.common.modules.chat_markers import ChatMarkers
from gajim.common.modules.chatstates import Chatstate
from gajim.common.modules.contacts import Contacts
from gajim.common.modules.discovery import Discovery
from gajim.common.modules.entity_time import EntityTime
from gajim.common.modules.gateway import Gateway
from gajim.common.modules.http_auth import HTTPAuth
from gajim.common.modules.http_muc_search import HttpMucSearch
from gajim.common.modules.httpupload import HTTPUpload
from gajim.common.modules.ibb import IBB
from gajim.common.modules.iq import Iq
from gajim.common.modules.jingle import Jingle
from gajim.common.modules.last_activity import LastActivity
from gajim.common.modules.mam import MAM
from gajim.common.modules.mds import MDS
from gajim.common.modules.message import Message
from gajim.common.modules.moderations import Moderations
from gajim.common.modules.muc import MUC
from gajim.common.modules.muc_blocking import MucBlocking
from gajim.common.modules.omemo import OMEMO
from gajim.common.modules.openpgp import OpenPGP
from gajim.common.modules.pep import PEP
from gajim.common.modules.ping import Ping
from gajim.common.modules.presence import Presence
from gajim.common.modules.pubsub import PubSub
from gajim.common.modules.reactions import Reactions
from gajim.common.modules.receipts import Receipts
from gajim.common.modules.register import Register
from gajim.common.modules.retraction import Retraction
from gajim.common.modules.roster import Roster
from gajim.common.modules.roster_item_exchange import RosterItemExchange
from gajim.common.modules.search import Search
from gajim.common.modules.security_labels import SecLabels
from gajim.common.modules.software_version import SoftwareVersion
from gajim.common.modules.user_avatar import UserAvatar
from gajim.common.modules.user_location import UserLocation
from gajim.common.modules.user_nickname import UserNickname
from gajim.common.modules.user_tune import UserTune
from gajim.common.modules.vcard4 import VCard4
from gajim.common.modules.vcard_avatars import VCardAvatars
from gajim.common.modules.vcard_temp import VCardTemp

if TYPE_CHECKING:
    from gajim.common.types import Client

ModulesT = (
    AdHocCommands
    | Annotations
    | BitsOfBinary
    | Blocking
    | Bookmarks
    | Bytestream
    | Caps
    | Carbons
    | ChatMarkers
    | Chatstate
    | Contacts
    | Discovery
    | EntityTime
    | Gateway
    | HTTPAuth
    | HttpMucSearch
    | HTTPUpload
    | IBB
    | Iq
    | Jingle
    | LastActivity
    | MAM
    | MDS
    | Message
    | Moderations
    | MUC
    | MucBlocking
    | OMEMO
    | OpenPGP
    | PEP
    | Ping
    | Presence
    | PubSub
    | Reactions
    | Receipts
    | Register
    | Retraction
    | Roster
    | RosterItemExchange
    | Search
    | SecLabels
    | SoftwareVersion
    | UserAvatar
    | UserLocation
    | UserNickname
    | UserTune
    | VCard4
    | VCardAvatars
    | VCardTemp
)

ModulesLiteralT = Literal[
    "AdHocCommands",
    "Annotations",
    "BitsOfBinary",
    "Blocking",
    "Bookmarks",
    "Bytestream",
    "Caps",
    "Carbons",
    "ChatMarkers",
    "Chatstate",
    "Contacts",
    "Discovery",
    "EntityTime",
    "Gateway",
    "HTTPAuth",
    "HttpMucSearch",
    "HTTPUpload",
    "IBB",
    "Iq",
    "Jingle",
    "LastActivity",
    "MAM",
    "MDS",
    "Message",
    "Moderations",
    "MUC",
    "MucBlocking",
    "OMEMO",
    "OpenPGP",
    "PEP",
    "Ping",
    "Presence",
    "PubSub",
    "Reactions",
    "Receipts",
    "Register",
    "Retraction",
    "Roster",
    "RosterItemExchange",
    "Search",
    "SecLabels",
    "SoftwareVersion",
    "UserAvatar",
    "UserLocation",
    "UserNickname",
    "UserTune",
    "VCard4",
    "VCardAvatars",
    "VCardTemp",
]


class ModuleDict(TypedDict):
    AdHocCommands: AdHocCommands
    Annotations: Annotations
    BitsOfBinary: BitsOfBinary
    Blocking: Blocking
    Bookmarks: Bookmarks
    Bytestream: Bytestream
    Caps: Caps
    Carbons: Carbons
    ChatMarkers: ChatMarkers
    Chatstate: Chatstate
    Contacts: Contacts
    Discovery: Discovery
    EntityTime: EntityTime
    Gateway: Gateway
    HTTPAuth: HTTPAuth
    HttpMucSearch: HttpMucSearch
    HTTPUpload: HTTPUpload
    IBB: IBB
    Iq: Iq
    Jingle: Jingle
    LastActivity: LastActivity
    MAM: MAM
    MDS: MDS
    Message: Message
    Moderations: Moderations
    MUC: MUC
    MucBlocking: MucBlocking
    OMEMO: OMEMO
    OpenPGP: OpenPGP
    PEP: PEP
    Ping: Ping
    Presence: Presence
    PubSub: PubSub
    Reactions: Reactions
    Receipts: Receipts
    Register: Register
    Retraction: Retraction
    Roster: Roster
    RosterItemExchange: RosterItemExchange
    Search: Search
    SecLabels: SecLabels
    SoftwareVersion: SoftwareVersion
    UserAvatar: UserAvatar
    UserLocation: UserLocation
    UserNickname: UserNickname
    UserTune: UserTune
    VCard4: VCard4
    VCardAvatars: VCardAvatars
    VCardTemp: VCardTemp


_modules: dict[str, ModuleDict] = {}
_store_publish_modules = (
    "UserLocation",
    "UserTune",
)


log = logging.getLogger("gajim.c.m")


def register_modules(client: "Client") -> None:
    if client.account in _modules:
        return

    _modules[client.account] = {}  # type: ignore

    for name, module_cls in ModuleDict.__annotations__.items():
        _modules[client.account][name] = module_cls.get_instance(client)


def register_single_module(client: "Client", instance: BaseModule, name: str) -> None:

    if client.account not in _modules:
        raise ValueError("Unknown account name: %s" % client.account)
    _modules[client.account][name] = instance


def unregister_modules(client: "Client") -> None:
    for instance in _modules[client.account].values():
        instance = cast(ModulesT, instance)
        if hasattr(instance, "cleanup"):
            instance.cleanup()
        app.check_finalize(instance)
    del _modules[client.account]


def unregister_single_module(client: "Client", name: str) -> None:
    if client.account not in _modules:
        return
    if name not in _modules[client.account]:
        return
    del _modules[client.account][name]


def send_stored_publish(account: str) -> None:
    for name in _store_publish_modules:
        _modules[account][name].send_stored_publish()


def get_module(account: str, name: ModulesLiteralT) -> ModulesT:
    return _modules[account][name]


def get_handlers(client: "Client") -> list[StanzaHandler]:
    handlers: list[StanzaHandler] = []
    for module in _modules[client.account].values():
        module = cast(ModulesT, module)
        handlers += module.handlers
    return handlers



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import TypeVar

from gi.repository import Gio
from gi.repository import GObject

_T = TypeVar("_T")


class SignalManager:
    def __init__(self) -> None:
        self._signal_data: list[tuple[GObject.Object | Gio.Action, int]] = []

    def _connect(
        self,
        obj: GObject.Object | Gio.Action,
        signal_name: str,
        callback: Any,
        *args: Any,
    ) -> int:
        signal_id = obj.connect(signal_name, callback, *args)
        self._signal_data.append((obj, signal_id))
        return signal_id

    def _connect_after(
        self, obj: GObject.Object, signal_name: str, callback: Any, *args: Any
    ) -> int:
        signal_id = obj.connect_after(signal_name, callback, *args)
        self._signal_data.append((obj, signal_id))
        return signal_id

    def _disconnect_all(self):
        for obj, signal_id in self._signal_data:
            obj.disconnect(signal_id)
        self._signal_data.clear()

    def _disconnect_object(self, obj: GObject.Object) -> None:
        for obj_, signal_id in list(self._signal_data):
            if obj is obj_:
                obj.disconnect(signal_id)
                self._signal_data.remove((obj, signal_id))



# Copyright (C) 2016-2018 Philipp Hörist <philipp AT hoerist.com>
# Copyright (C) 2005-2006 Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2005-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2008 Stephan Erb <steve-e AT h3c.de>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing
from typing import Any

import json
import platform
import pprint
import sys
import threading
import traceback
import webbrowser
from io import StringIO
from types import TracebackType
from urllib.parse import urlencode

import nbxmpp
from gi.repository import Gtk

import gajim
from gajim.common import app
from gajim.common.file_transfer_manager import FileTransfer
from gajim.common.helpers import determine_proxy
from gajim.common.i18n import _
from gajim.common.util.version import get_glib_version
from gajim.common.util.version import get_gobject_version
from gajim.common.util.version import get_os_name
from gajim.common.util.version import get_os_version
from gajim.common.util.version import get_soup_version

from gajim.gtk.builder import get_builder
from gajim.gtk.util.misc import get_adw_version
from gajim.gtk.util.misc import get_gtk_version
from gajim.gtk.window import GajimAppWindow

try:
    import sentry_sdk
except Exception:
    # Sentry has a lot of side effects on import
    # make sure this optional dependency does not prevent
    # Gajim from starting
    if typing.TYPE_CHECKING:
        import sentry_sdk

_exception_in_progress = threading.Lock()

ISSUE_URL = "https://dev.gajim.org/gajim/gajim/issues/new"

ISSUE_TEXT = """## Versions
- OS: {}
- GTK: {}
- PyGObject: {}
- GLib: {}
- Adw: {}
- libsoup: {}
- python-nbxmpp: {}
- Gajim: {}

## Traceback
```
{}
```

## Steps to reproduce the problem

  1.
  1.
  1.

## Expected behavior
...

## Actual behavior
...
"""


def _hook(type_: type[BaseException], value: BaseException, tb: TracebackType) -> None:
    if not _exception_in_progress.acquire(False):
        # Exceptions have piled up, so we use the default exception
        # handler for such exceptions
        sys.__excepthook__(type_, value, tb)
        return

    ExceptionDialog(type_, value, tb)


class ExceptionDialog(GajimAppWindow):
    def __init__(
        self, type_: type[BaseException], value: BaseException, tb: TracebackType
    ) -> None:
        GajimAppWindow.__init__(
            self,
            name="ExceptionDialog",
            title=_("Gajim - Error"),
            default_width=700,
            header_bar=True,
        )

        self._traceback_data = (type_, value, tb)
        self._sentry_available = app.is_installed("SENTRY_SDK")

        self._ui = get_builder("exception_dialog.ui")
        self.set_child(self._ui.exception_box)

        if not self._sentry_available:
            self._ui.user_feedback_box.set_visible(False)
            self._ui.infobar.set_reveal_child(True)

        self._ui.report_button.grab_focus()
        self.set_default_widget(self._ui.report_button)

        self._connect(self._ui.close_button, "clicked", self._on_close_clicked)
        self._connect(self._ui.report_button, "clicked", self._on_report_clicked)

        trace = StringIO()
        traceback.print_exception(type_, value, tb, None, trace)

        self._issue_text = self._get_issue_text(trace.getvalue())
        buffer_ = self._ui.exception_view.get_buffer()
        buffer_.set_text(self._issue_text)

        if self._sentry_available:
            self._ui.user_feedback_entry.grab_focus()

        self.show()

    def _on_report_clicked(self, _button: Gtk.Button) -> None:
        if self._sentry_available and determine_proxy() is None:
            # sentry-sdk supports a http-proxy arg but for now only use
            # sentry when no proxy is set, because we never tested if this
            # works. It's not worth it to potentially leak users identity just
            # because of error reporting.
            self._report_with_sentry()
        else:
            self._report_with_browser()

    def _report_with_browser(self):
        params = {"issue[description]": self._issue_text}
        url = f"{ISSUE_URL}?{urlencode(params)}"
        webbrowser.open(url, new=2)
        self.close()

    def _on_close_clicked(self, _button: Gtk.Button) -> None:
        self.close()

    @staticmethod
    def _get_issue_text(traceback_text: str) -> str:
        return ISSUE_TEXT.format(
            f"{get_os_name()} {get_os_version()}",
            get_gtk_version(),
            get_gobject_version(),
            get_glib_version(),
            get_adw_version(),
            get_soup_version(),
            nbxmpp.__version__,
            gajim.__version__,
            traceback_text,
        )

    def _report_with_sentry(self) -> None:
        if sentry_sdk.last_event_id() is None:
            # Sentry has not been initialized yet:
            # update sentry endpoint, init sentry, then capture exception
            self._request_sentry_endpoint()
            return

        self._capture_exception()
        self.close()

    def _request_sentry_endpoint(self) -> None:
        self._ui.report_button.set_sensitive(False)
        self._ui.close_button.set_sensitive(False)
        self._ui.report_spinner.set_visible(True)

        app.ftm.http_request(
            "GET",
            "https://gajim.org/updates.json",
            callback=self._on_endpoint_received,
        )

    def _parse_endpoint(self, content: bytes) -> str:
        try:
            data = json.loads(content)
        except Exception as error:
            raise ValueError(f"Json parsing error: {error}")

        endpoint = data.get("sentry_endpoint")
        if endpoint is None:
            raise ValueError("Sentry endpoint missing in response")

        return endpoint

    def _on_endpoint_received(self, obj: FileTransfer) -> None:
        try:
            result = obj.get_result()
            endpoint = self._parse_endpoint(result.content)
        except Exception as error:
            print("Failed to retrieve sentry endpoint: %s" % error)
            self._report_with_browser()

        else:
            self._init_sentry(endpoint)
            self._capture_exception()
            self.close()

    def _init_sentry(self, endpoint: str) -> None:
        sentry_sdk.init(
            dsn=endpoint,
            traces_sample_rate=0.0,
            max_breadcrumbs=0,
            release=gajim.__version__,
            default_integrations=False,
            shutdown_timeout=0,
            auto_session_tracking=False,
            before_send=self._before_send,  # type: ignore
            debug=False,
        )

        sentry_sdk.set_context(
            "os", {"name": get_os_name(), "version": get_os_version()}
        )

        sentry_sdk.set_context(
            "software",
            {
                "python": platform.python_version(),
                "python-nbxmpp": nbxmpp.__version__,
                "GTK": get_gtk_version(),
                "PyGObject": get_gobject_version(),
                "GLib": get_glib_version(),
                "Adw": get_adw_version(),
                "libsoup": get_soup_version(),
            },
        )

    def _capture_exception(self) -> None:
        sentry_sdk.set_context(
            "user_feedback", {"Feedback": self._ui.user_feedback_entry.get_text()}
        )
        sentry_sdk.capture_exception(self._traceback_data)

    def _before_send(self, event: dict[str, Any], hint: Any) -> dict[str, Any]:
        # Make sure the exception value is set, GitLab needs it.
        # The value is the arg which is passed to the Exception.
        # e.g. raise Exception('Error')
        try:
            value = event["exception"]["values"][0].get("value")
            if not value:
                event["exception"]["values"][0]["value"] = "Unknown"
        except Exception:
            pass

        # Remove the hostname of the machine
        event["server_name"] = "redacted"
        pprint.pprint(event)
        return event

    def _cleanup(self) -> None:
        _exception_in_progress.release()


def init() -> None:
    if sys.platform == "win32" or not sys.stderr.isatty():
        sys.excepthook = _hook



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app
from gajim.common import ged
from gajim.common.const import AvatarSize
from gajim.common.events import AccountDisabled
from gajim.common.events import AccountEnabled
from gajim.common.events import RegisterActions
from gajim.common.ged import EventHelper
from gajim.common.i18n import _

from gajim.gtk.chat_list_row import ChatListRow
from gajim.gtk.chat_page import ChatPage
from gajim.gtk.menus import get_activity_feed_menu
from gajim.gtk.sidebar_listbox import SideBarListBox
from gajim.gtk.sidebar_listbox import SideBarListBoxRow
from gajim.gtk.status_selector import StatusSelectorPopover
from gajim.gtk.util.misc import get_ui_string
from gajim.gtk.widgets import GajimPopover
from gajim.gtk.workspace_listbox import WorkspaceListBox


@Gtk.Template(string=get_ui_string("app_side_bar.ui"))
class AppSideBar(Gtk.Box, EventHelper):
    __gtype_name__ = "AppSideBar"

    _top_listbox: SideBarListBox = Gtk.Template.Child()
    _activity_row: SideBarListBoxRow = Gtk.Template.Child()
    _workspace_listbox: WorkspaceListBox = Gtk.Template.Child()
    _bottom_listbox: SideBarListBox = Gtk.Template.Child()
    _toggle_row: SideBarListBoxRow = Gtk.Template.Child()
    _account_row: SideBarListBoxRow = Gtk.Template.Child()

    def __init__(self) -> None:
        Gtk.Box.__init__(self)
        EventHelper.__init__(self)

        self._activity_feed_popover_menu = GajimPopover(menu=get_activity_feed_menu())
        self.append(self._activity_feed_popover_menu)

        self._account_popover = AccountPopover()
        self._account_popover.set_parent(self._account_row)
        self._account_popover.connect("clicked", self._on_account_clicked)

        self._status_popover = StatusSelectorPopover()
        self._status_popover.set_parent(self._account_row)
        self._status_popover.connect("clicked", self._on_status_clicked)

        accounts = app.settings.get_active_accounts()
        for account in reversed(accounts):
            client = app.get_client(account)
            contact = client.get_own_contact()

            client.connect_signal("state-changed", self._update_account_row)
            contact.connect("avatar-update", self._update_account_row)

        self._update_account_row()

        self.register_events(
            [
                ("account-enabled", ged.GUI1, self._on_account_enabled),
                ("account-disabled", ged.GUI1, self._on_account_disabled),
                ("our-show", ged.GUI1, self._update_account_row),
                ("register-actions", ged.GUI1, self._on_register_actions),
            ]
        )

    def _on_register_actions(self, _event: RegisterActions) -> None:
        action = app.window.lookup_action("chat-list-visible")
        assert action is not None
        action.bind_property(
            "state",
            self._toggle_row,
            "icon-name",
            GObject.BindingFlags.SYNC_CREATE,
            transform_to=self._transform_to_icon_name,
        )
        action.bind_property(
            "state",
            self._toggle_row,
            "tooltip-text",
            GObject.BindingFlags.SYNC_CREATE,
            transform_to=self._transform_to_tooltip_text,
        )

    def _on_account_enabled(self, event: AccountEnabled) -> None:
        client = app.get_client(event.account)
        contact = client.get_own_contact()

        client.connect_signal("state-changed", self._update_account_row)
        contact.connect("avatar-update", self._update_account_row)

        self._update_account_row()

    def _on_account_disabled(self, event: AccountDisabled) -> None:
        client = app.get_client(event.account)
        contact = client.get_own_contact()

        client.disconnect_all_from_obj(self)
        contact.disconnect_all_from_obj(self)

        self._update_account_row()

    def _update_account_row(self, *args: Any) -> None:
        accounts = app.settings.get_active_accounts()
        if len(accounts) == 1:
            account = accounts[0]
        else:
            account = None

        texture = app.app.avatar_storage.get_account_button_texture(
            account,
            AvatarSize.ACCOUNT_SIDE_BAR,
            self.get_scale_factor(),
            self._has_connectivity_problems(),
        )
        self._account_row.set_from_paintable(texture)

    @staticmethod
    def _has_connectivity_problems() -> bool:
        for account in app.settings.get_active_accounts():
            client = app.get_client(account)
            if client.has_connectivity_problems():
                return True
        return False

    def _on_account_clicked(
        self,
        _popover: AccountPopover,
        account: str,
    ) -> None:
        app.window.show_account_page(account)
        self._bottom_listbox.select_row(self._account_row)

    def _on_status_clicked(self, _button: Gtk.Button, status: str) -> None:
        accounts = app.get_connected_accounts()
        account = accounts[0] if len(accounts) == 1 else None
        app.app.change_status(status=status, account=account)

    @Gtk.Template.Callback()
    def _on_account_button_press(
        self,
        gesture_click: Gtk.GestureClick,
        _n_press: int,
        x: float,
        y: float,
    ) -> None:
        current_button = gesture_click.get_current_button()
        if current_button == Gdk.BUTTON_PRIMARY:
            # Left click
            gesture_click.set_state(Gtk.EventSequenceState.CLAIMED)
            self._status_popover.popdown()

            accounts = app.settings.get_active_accounts()
            if len(accounts) == 0:
                return

            if len(accounts) == 1:
                app.window.show_account_page(accounts[0])
                self._bottom_listbox.select_row(self._account_row)
            else:
                self._account_popover.popup()

        if current_button == Gdk.BUTTON_SECONDARY:
            # Right click
            gesture_click.set_state(Gtk.EventSequenceState.CLAIMED)
            self._status_popover.popup()
            self._account_popover.popdown()

    @Gtk.Template.Callback()
    def _on_activity_feed_button_press(
        self,
        gesture_click: Gtk.GestureClick,
        _n_press: int,
        x: float,
        y: float,
    ) -> None:
        current_button = gesture_click.get_current_button()
        if current_button == Gdk.BUTTON_PRIMARY:
            # Left click
            gesture_click.set_state(Gtk.EventSequenceState.CLAIMED)
            self._activity_row.activate()
            app.window.show_activity_page()

        if current_button == Gdk.BUTTON_SECONDARY:
            # Right click
            gesture_click.set_state(Gtk.EventSequenceState.CLAIMED)
            self._activity_feed_popover_menu.set_pointing_to_coord(x, y)
            self._activity_feed_popover_menu.popup()

    @staticmethod
    def _transform_to_icon_name(
        binding: GObject.Binding, is_visible: GLib.Variant
    ) -> str:
        direction = "right" if is_visible.unpack() else "left"
        return f"lucide-chevron-{direction}-symbolic"

    @staticmethod
    def _transform_to_tooltip_text(
        binding: GObject.Binding, is_visible: GLib.Variant
    ) -> str:
        if is_visible.unpack():
            return _("Hide Chat List")
        return _("Show Chat List")

    def set_chat_page(self, chat_page: ChatPage) -> None:
        self._activity_row.set_unread_notify(chat_page.get_activity_list())
        self._workspace_listbox.set_chat_page(chat_page)

    def select_chat(self) -> None:
        self._top_listbox.unselect_all()
        self._activity_row.set_selectable(False)
        self._bottom_listbox.unselect_all()

    def show_account_page(self) -> None:
        self._top_listbox.unselect_all()
        self._activity_row.set_selectable(False)
        self._workspace_listbox.unselect_all()

    def show_activity_page(self) -> None:
        self._activity_row.set_selectable(True)
        self._top_listbox.select_row(self._activity_row)
        self._bottom_listbox.unselect_all()
        self._workspace_listbox.unselect_all()

    def activate_workspace(self, workspace_id: str) -> None:
        self._top_listbox.unselect_all()
        self._activity_row.set_selectable(False)
        self._bottom_listbox.unselect_all()
        self._workspace_listbox.activate_workspace(workspace_id)

    def get_active_workspace(self) -> str | None:
        return self._workspace_listbox.get_active_workspace()

    def highlight_dnd_targets(self, dragged_object: Any, highlight: bool) -> None:
        if not isinstance(dragged_object, SideBarListBoxRow | ChatListRow):
            return

        if highlight:
            self._workspace_listbox.add_css_class("dnd-target")
        else:
            self._workspace_listbox.remove_css_class("dnd-target")

    def add_workspace(self, workspace_id: str) -> None:
        self._workspace_listbox.add_workspace(workspace_id)

    def remove_workspace(self, workspace_id: str) -> None:
        self._workspace_listbox.remove_workspace(workspace_id)

    def get_other_workspace(self, workspace_id: str) -> str | None:
        return self._workspace_listbox.get_other_workspace(workspace_id)

    def store_workspace_order(self) -> None:
        self._workspace_listbox.store_workspace_order()

    def update_workspace(self, workspace_id: str) -> None:
        self._workspace_listbox.update_avatar(workspace_id)

    def get_first_workspace(self) -> str:
        return self._workspace_listbox.get_first_workspace()

    def activate_workspace_number(self, number: int) -> None:
        self._workspace_listbox.activate_workspace_number(number)


@Gtk.Template(string=get_ui_string("account_popover.ui"))
class AccountPopover(Gtk.Popover):
    __gtype_name__ = "AccountPopover"

    __gsignals__ = {
        "clicked": (
            GObject.SignalFlags.RUN_LAST,
            None,
            (str,),
        )
    }

    _listbox: Gtk.ListBox = Gtk.Template.Child()

    def __init__(self) -> None:
        Gtk.Popover.__init__(self)
        self._rows: list[AccountPopoverRow] = []

        self.connect("closed", self._on_closed)

    def popup(self) -> None:
        for account in app.settings.get_active_accounts():
            row = AccountPopoverRow(account=account)
            self._listbox.append(row)
            self._rows.append(row)

        Gtk.Popover.popup(self)

    def _on_closed(self, popover: Gtk.Popover) -> None:
        for row in self._rows:
            self._listbox.remove(row)
        self._rows.clear()

    @Gtk.Template.Callback()
    def _on_row_activated(self, _listbox: Gtk.ListBox, row: AccountPopoverRow) -> None:
        account = row.get_account()

        self.emit("clicked", account)
        self.popdown()


@Gtk.Template(string=get_ui_string("account_popover_row.ui"))
class AccountPopoverRow(Gtk.ListBoxRow):
    __gtype_name__ = "AccountPopoverRow"

    _color_bar: Gtk.Box = Gtk.Template.Child()
    _avatar: Gtk.Image = Gtk.Template.Child()
    _label: Gtk.Label = Gtk.Template.Child()

    def __init__(self, account: str = "") -> None:
        Gtk.ListBoxRow.__init__(self)

        self._account = account

        color_class = app.css_config.get_dynamic_class(account)
        self._color_bar.add_css_class(color_class)

        label = app.settings.get_account_setting(account, "account_label")
        self._label.set_text(label)

        client = app.get_client(account)
        connectivity_problems = client.has_connectivity_problems()
        texture = app.app.avatar_storage.get_account_button_texture(
            account,
            AvatarSize.ACCOUNT_SIDE_BAR,
            self.get_scale_factor(),
            connectivity_problems,
        )
        self._avatar.set_from_paintable(texture)
        if connectivity_problems:
            self._avatar.set_tooltip_text(_("There are connectivity issues"))
        else:
            self._avatar.set_tooltip_text("")

    def do_unroot(self) -> None:
        Gtk.ListBoxRow.do_unroot(self)
        app.check_finalize(self)

    def get_account(self) -> str:
        return self._account



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any
from typing import cast

import logging

from gi.repository import Gio
from gi.repository import GLib

from gajim.common import app
from gajim.common import events
from gajim.common import ged
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.structs import OutgoingMessage
from gajim.common.util.status import get_client_status
from gajim.common.util.status import get_global_show
from gajim.common.util.status import get_global_status_message
from gajim.common.util.version import gi_package_version

log = logging.getLogger("gajim.c.dbus.remote_control")

INTERFACE_DESC = """
<!DOCTYPE node PUBLIC '-//freedesktop//DTD D-BUS Object Introspection 1.0//EN'
'http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd'>
<node>
    <interface name='org.freedesktop.DBus.Introspectable'>
        <method name='Introspect'>
            <arg name='data' direction='out' type='s'/>
        </method>
    </interface>
    <interface name='org.gajim.dbus.RemoteInterface'>
        <method name='account_info'>
            <arg name='account' type='s' />
            <arg direction='out' type='a{ss}' />
        </method>
        <method name='change_status'>
            <arg name='status' type='s' />
            <arg name='message' type='s' />
            <arg name='account' type='s' />
            <arg direction='out' type='b' />
        </method>
        <method name='get_status'>
            <arg name='account' type='s' />
            <arg direction='out' type='s' />
        </method>
        <method name='get_status_message'>
            <arg name='account' type='s' />
            <arg direction='out' type='s' />
        </method>
        <method name='get_unread_msgs_number'>
            <arg direction='out' type='s' />
        </method>
        <method name='list_accounts'>
            <arg direction='out' type='as' />
        </method>
        <method name='list_contacts'>
            <arg name='account' type='s' />
            <arg direction='out' type='aa{sv}' />
        </method>
        <method name='send_chat_message'>
            <arg name='jid' type='s' />
            <arg name='message' type='s' />
            <arg name='account' type='s' />
            <arg direction='out' type='b' />
        </method>
        <method name='send_groupchat_message'>
            <arg name='room_jid' type='s' />
            <arg name='message' type='s' />
            <arg name='account' type='s' />
            <arg direction='out' type='b' />
        </method>
        <signal name='AccountPresence'>
            <arg type='av' />
        </signal>
        <signal name='ContactPresence'>
            <arg type='av' />
        </signal>
        <signal name='GCMessage'>
            <arg type='av' />
        </signal>
        <signal name='MessageSent'>
            <arg type='av' />
        </signal>
        <signal name='NewMessage'>
            <arg type='av' />
        </signal>
    </interface>
</node>
"""


def get_dbus_struct(obj: Any) -> GLib.Variant:
    """
    Recursively go through all the items and replace them with their casted dbus
    equivalents
    """
    if isinstance(obj, str):
        return GLib.Variant("s", obj)
    if isinstance(obj, int):
        return GLib.Variant("i", obj)
    if isinstance(obj, float):
        return GLib.Variant("d", obj)
    if isinstance(obj, bool):
        return GLib.Variant("b", obj)
    if isinstance(obj, list | tuple):
        lst = [get_dbus_struct(i) for i in obj if i is not None]  # type: ignore
        result = GLib.Variant("av", lst)
        return result
    if isinstance(obj, dict):
        obj = cast(dict[str, Any], obj)
        result = GLib.VariantDict()
        for key, value in obj.items():
            result.insert_value(key, get_dbus_struct(value))
        return result.end()
    # unknown type
    return GLib.Variant("s", str(obj))


class Server:
    def __init__(self, con: Gio.DBusConnection, path: str) -> None:
        self._method_outargs: dict[str, str] = {}
        self._method_inargs: dict[str, tuple[str, ...]] = {}
        node_info = Gio.DBusNodeInfo.new_for_xml(INTERFACE_DESC)
        for interface in node_info.interfaces:
            for method in interface.methods:
                self._method_outargs[method.name] = (
                    "(" + "".join([arg.signature for arg in method.out_args]) + ")"
                )
                self._method_inargs[method.name] = tuple(
                    arg.signature for arg in method.in_args
                )

            if gi_package_version("GLib>=2.84.0"):
                register_method = con.register_object_with_closures2
            else:
                register_method = con.register_object

            register_method(
                object_path=path,
                interface_info=interface,
                method_call_closure=self._on_method_call,
            )

    def _on_method_call(
        self,
        _connection: Gio.DBusConnection,
        _sender: str,
        _object_path: str,
        _interface_name: str,
        method_name: str,
        parameters: GLib.Variant,
        invocation: Gio.DBusMethodInvocation,
    ) -> None:
        args = list(parameters.unpack())
        for i, sig in enumerate(self._method_inargs[method_name]):
            if sig == "h":
                msg = invocation.get_message()
                fd_list = msg.get_unix_fd_list()
                assert fd_list is not None
                args[i] = fd_list.get(args[i])

        result = getattr(self, method_name)(*args)

        # out_args is at least (signature1). We therefore always wrap the result
        # as a tuple. Refer to https://bugzilla.gnome.org/show_bug.cgi?id=765603
        result = (result,)

        out_args = self._method_outargs[method_name]
        if out_args != "()":
            variant = GLib.Variant(out_args, result)
            invocation.return_value(variant)
        else:
            invocation.return_value(None)


class GajimRemote(Server):
    def __init__(self) -> None:
        application_id = app.app.get_application_id()
        assert application_id is not None
        self._con = Gio.bus_get_sync(Gio.BusType.SESSION, None)
        Gio.bus_own_name_on_connection(
            self._con, application_id, Gio.BusNameOwnerFlags.NONE, None, None
        )
        super().__init__(self._con, "/org/gajim/dbus/RemoteObject")

        app.ged.register_event_handler(
            "presence-received", ged.POSTGUI, self._on_presence_received
        )
        app.ged.register_event_handler(
            "message-received", ged.POSTGUI, self._on_message_received
        )
        app.ged.register_event_handler("our-show", ged.POSTGUI, self._on_our_status)
        app.ged.register_event_handler(
            "message-sent", ged.POSTGUI, self._on_message_sent
        )

    def _on_message_sent(self, event: events.MessageSent) -> None:
        message = event.message
        self.raise_signal("MessageSent", (event.account, [event.jid, message.text]))

    def _on_presence_received(self, event: events.PresenceReceived) -> None:
        self.raise_signal(
            "ContactPresence",
            (event.account, [event.jid, event.resource, event.show, event.status]),
        )

    def _on_message_received(self, event: events.MessageReceived) -> None:
        message = event.message
        self.raise_signal(
            "NewMessage",
            (
                event.account,
                [
                    message.remote.jid,
                    message.resource,
                    message.type,
                    message.timestamp,
                    message.text,
                ],
            ),
        )

    def _on_our_status(self, event: events.ShowChanged) -> None:
        self.raise_signal("AccountPresence", (event.show, event.account))

    def raise_signal(self, event_name: str, data: Any) -> None:
        log.info("Send event %s", event_name)
        self._con.emit_signal(
            None,
            "/org/gajim/dbus/RemoteObject",
            "org.gajim.dbus.RemoteInterface",
            event_name,
            GLib.Variant.new_tuple(get_dbus_struct(data)),
        )

    @staticmethod
    def get_status(account: str) -> str:
        if not account:
            return get_global_show()
        return get_client_status(account)

    @staticmethod
    def get_status_message(account: str) -> str:
        if not account:
            return get_global_status_message()
        return app.get_client(account).status_message

    @staticmethod
    def _send_message(jid: str, message: str, account: str, type_: str) -> bool:
        if not app.account_is_available(account):
            return False

        client = app.get_client(account)
        contact = client.get_module("Contacts").get_contact(
            jid, groupchat=type_ == "groupchat"
        )

        if isinstance(contact, GroupchatContact):
            if not contact.is_joined:
                return False

        assert isinstance(
            contact, BareContact | GroupchatContact | GroupchatParticipant
        )
        message_ = OutgoingMessage(account=account, contact=contact, text=message)

        app.get_client(account).send_message(message_)
        return True

    def send_chat_message(self, jid: str, message: str, account: str) -> bool:
        if not jid or not message or not account:
            return False

        return self._send_message(jid, message, account, "chat")

    def send_groupchat_message(self, jid: str, message: str, account: str) -> bool:
        if not jid or not message or not account:
            return False

        return self._send_message(jid, message, account, "groupchat")

    @staticmethod
    def change_status(status: str, message: str, account: str) -> bool:
        """
        change_status(status, message, account). Account is optional - if not
        specified status is changed for all accounts
        """
        if status not in ("offline", "online", "chat", "away", "xa", "dnd"):
            status = ""

        if account:
            if not status:
                if account not in app.settings.get_active_accounts():
                    return False
                status = app.get_client(account).status

            GLib.idle_add(app.get_client(account).change_status, status, message)
        else:
            # account not specified, so change the status of all accounts
            for acc in app.settings.get_active_accounts():
                if not app.settings.get_account_setting(acc, "sync_with_global_status"):
                    continue

                if not status:
                    status = app.get_client(acc).status

                GLib.idle_add(app.get_client(acc).change_status, status, message)
        return True

    @staticmethod
    def list_accounts() -> list[str]:
        """
        List register accounts
        """
        result = app.settings.get_active_accounts()
        result_array: list[str] = []
        if result:
            for account in result:
                result_array.append(account)
        return result_array

    @staticmethod
    def account_info(account: str) -> dict[str, str]:
        """
        Show info on account: resource, jid, nick, message
        """
        result: dict[str, str] = {}
        if account in app.settings.get_active_accounts():
            # account is valid
            client = app.get_client(account)
            result["status"] = client.status
            result["name"] = client.name
            result["jid"] = app.get_jid_from_account(client.name)
            result["message"] = client.status_message
            result["resource"] = app.settings.get_account_setting(
                client.name, "resource"
            )
        return result

    def list_contacts(self, account: str) -> list[dict[str, GLib.Variant]]:
        result: list[dict[str, GLib.Variant]] = []
        accounts = app.settings.get_active_accounts()
        if not accounts:
            return result
        if account:
            accounts_to_search = [account]
        else:
            accounts_to_search = accounts
        for acct in accounts_to_search:
            if acct in accounts:
                client = app.get_client(acct)
                for contact in client.get_module("Roster").iter_contacts():
                    item = self._contacts_as_dbus_structure(contact)
                    if item:
                        result.append(item)
        return result

    @staticmethod
    def _contacts_as_dbus_structure(contact: BareContact) -> dict[str, GLib.Variant]:
        """
        Get info from list of Contact objects and create dbus dict
        """

        contact_dict: dict[str, GLib.Variant] = {}

        contact_dict["name"] = GLib.Variant("s", contact.name)
        contact_dict["show"] = GLib.Variant("s", contact.show.value)
        contact_dict["jid"] = GLib.Variant("s", str(contact.jid))

        resources = GLib.VariantBuilder(GLib.VariantType("a(ss)"))
        for res_contact in contact.iter_resources():
            resource_props = (res_contact.resource, res_contact.status)
            resources.add_value(GLib.Variant("(ss)", resource_props))
        contact_dict["resources"] = resources.end()

        groups = GLib.VariantBuilder(GLib.VariantType("as"))
        for group in contact.groups:
            groups.add_value(GLib.Variant("s", group))
        contact_dict["groups"] = groups.end()
        return contact_dict

    @staticmethod
    def get_unread_msgs_number() -> str:
        return str(app.window.get_total_unread_count())

    @staticmethod
    def Introspect() -> str:  # pylint: disable=invalid-name
        return INTERFACE_DESC



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import Gtk
from gi.repository import Pango
from nbxmpp.const import Chatstate

from gajim.common import app
from gajim.common import types
from gajim.common.i18n import _
from gajim.common.modules.contacts import GroupchatContact


class ChatStateIndicator(Gtk.Box):
    def __init__(self) -> None:
        Gtk.Box.__init__(
            self,
            spacing=20,
            margin_start=14,
        )

        self._contact: types.ChatContactT | None = None

        self._composing_icon = Gtk.Image.new_from_icon_name("lucide-ellipsis-symbolic")
        self._composing_icon.set_visible(False)
        self._composing_icon.add_css_class("chat-state-icon")
        self.append(self._composing_icon)

        self._label = Gtk.Label(
            max_width_chars=52,
            ellipsize=Pango.EllipsizeMode.END,
        )
        self._label.add_css_class("dimmed")
        self.append(self._label)

    def do_unroot(self) -> None:
        Gtk.Box.do_unroot(self)
        if self._contact is not None:
            self._contact.disconnect_all_from_obj(self)
        app.check_finalize(self)

    def switch_contact(self, contact: types.ChatContactT) -> None:
        if self._contact is not None:
            self._contact.disconnect_all_from_obj(self)

        self._contact = contact
        self._contact.connect("chatstate-update", self._on_chatstate_update)

        self._update_state()

    def _on_chatstate_update(
        self, _contact: types.BareContact, _signal_name: str
    ) -> None:
        self._update_state()

    def _update_state(self) -> None:
        assert self._contact is not None

        if isinstance(self._contact, GroupchatContact):
            visible = bool(self._contact.get_composers())
            text = self._get_muc_composing_text()

            self._composing_icon.add_css_class("chat-state-icon")
        else:
            visible = self._contact.chatstate in (Chatstate.COMPOSING, Chatstate.PAUSED)
            text = f"{self._contact.name} {self._contact.chatstate_string}"

            if self._contact.chatstate == Chatstate.COMPOSING:
                self._composing_icon.add_css_class("chat-state-icon")
            else:
                self._composing_icon.remove_css_class("chat-state-icon")

        self._composing_icon.set_visible(visible)
        self._label.set_text(text if visible else "")

    def _get_muc_composing_text(self) -> str:
        assert isinstance(self._contact, GroupchatContact)
        composers = tuple(c.name for c in self._contact.get_composers())
        count = len(composers)
        if count == 1:
            return _("%s is typing…") % composers[0]
        if count == 2:
            return _("%s and %s are typing…") % composers

        return _("%s participants are typing…") % count



# Copyright (C) 2018 Emmanuel Gil Peyrot <linkmauve AT linkmauve.fr>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import hashlib
import logging
from base64 import b64decode
from collections.abc import Callable
from pathlib import Path

import nbxmpp
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.structs import BobData
from nbxmpp.structs import IqProperties
from nbxmpp.structs import StanzaHandler

from gajim.common import app
from gajim.common import configpaths
from gajim.common import types
from gajim.common.modules.base import BaseModule

log = logging.getLogger("gajim.c.m.bob")


class BitsOfBinary(BaseModule):
    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.handlers = [
            StanzaHandler(
                name="iq",
                callback=self._answer_bob_request,
                typ="get",
                ns=Namespace.BOB,
            ),
        ]

        # Used to track which cids are in-flight.
        self.awaiting_cids: dict[str, list[tuple[Callable[..., Any], Any, int]]] = {}

    def _answer_bob_request(
        self, _con: types.NBXMPPClient, stanza: Iq, _properties: IqProperties
    ) -> None:
        self._log.info("Request from %s for BoB data", stanza.getFrom())
        iq = stanza.buildReply("error")
        err = nbxmpp.ErrorNode(nbxmpp.ERR_ITEM_NOT_FOUND)
        iq.addChild(node=err)
        self._log.info("Sending item-not-found")
        self._con.connection.send(iq)
        raise nbxmpp.NodeProcessed

    def _on_bob_received(
        self, _nbxmpp_client: types.NBXMPPClient, result: Iq, cid: str
    ) -> None:
        """
        Called when we receive BoB data
        """
        if cid not in self.awaiting_cids:
            return

        # pylint: disable=cell-var-from-loop
        if result.getType() == "result":
            data = result.getTags("data", namespace=Namespace.BOB)
            if data.getAttr("cid") == cid:
                for func in self.awaiting_cids[cid]:
                    cb = func[0]
                    args = func[1]
                    pos = func[2]
                    bob_data = data.getData()

                    def recurs(node: nbxmpp.Node, cid: str, data: BobData) -> None:
                        if node.getData() == "cid:" + cid:
                            node.setData(data)
                        else:
                            for child in node.getChildren():
                                recurs(child, cid, data)

                    recurs(args[pos], cid, bob_data)
                    cb(*args)
                del self.awaiting_cids[cid]
                return

        # An error occurred, call callback without modifying data.
        for func in self.awaiting_cids[cid]:
            cb = func[0]
            args = func[1]
            cb(*args)
        del self.awaiting_cids[cid]

    def get_bob_data(
        self, cid: str, to: str, callback: Callable[..., Any], args: Any, position: int
    ) -> None:
        """
        Request for BoB (XEP-0231) and when data will arrive, call callback
        with given args, after having replaced cid by it's data in
        args[position]
        """
        if cid in self.awaiting_cids:
            self.awaiting_cids[cid].append((callback, args, position))
        else:
            self.awaiting_cids[cid] = [(callback, args, position)]
        iq = nbxmpp.Iq(to=to, typ="get")
        iq.addChild(name="data", attrs={"cid": cid}, namespace=Namespace.BOB)
        self._con.connection.SendAndCallForResponse(
            iq, self._on_bob_received, {"cid": cid}
        )


def parse_bob_data(stanza: Iq) -> Path | None:
    data_node = stanza.getTag("data", namespace=Namespace.BOB)
    if data_node is None:
        return None

    cid = data_node.getAttr("cid")
    type_ = data_node.getAttr("type")
    max_age = data_node.getAttr("max-age")
    if max_age is not None:
        try:
            max_age = int(max_age)
        except Exception:
            log.exception(stanza)
            return None

    if cid is None or type_ is None:
        log.warning("Invalid data node (no cid or type attr): %s", stanza)
        return None

    try:
        algo_hash = cid.split("@")[0]
        algo, hash_ = algo_hash.split("+")
    except Exception:
        log.exception("Invalid cid: %s", stanza)
        return None

    bob_data = data_node.getData()
    if not bob_data:
        log.warning("No data found: %s", stanza)
        return None

    filepath = configpaths.get("BOB") / algo_hash
    if algo_hash in app.bob_cache or filepath.exists():
        log.info("BoB data already cached")
        return None

    try:
        bob_data = b64decode(bob_data)
    except Exception:
        log.warning("Unable to decode data")
        log.exception(stanza)
        return None

    if len(bob_data) > 10000:
        log.warning("%s: data > 10000 bytes", stanza.getFrom())
        return None

    try:
        sha = hashlib.new(algo)
    except ValueError as error:
        log.warning(stanza)
        log.warning(error)
        return None

    sha.update(bob_data)
    if sha.hexdigest() != hash_:
        log.warning("Invalid hash: %s", stanza)
        return None

    if max_age == 0:
        app.bob_cache[algo_hash] = bob_data
    else:
        try:
            with open(str(filepath), "w+b") as file:
                file.write(bob_data)
        except Exception:
            log.warning("Unable to save data")
            log.exception(stanza)
            return None

    log.info("BoB data stored: %s", algo_hash)
    return filepath


def store_bob_data(bob_data: BobData | None) -> Path | None:
    if bob_data is None:
        return None

    algo_hash = f"{bob_data.algo}+{bob_data.hash_}"

    filepath = configpaths.get("BOB") / algo_hash
    if algo_hash in app.bob_cache or filepath.exists():
        log.info("BoB data already cached")
        return None

    if bob_data.max_age == 0:
        app.bob_cache[algo_hash] = bob_data.data
    else:
        try:
            with open(str(filepath), "w+b") as file:
                file.write(bob_data.data)
        except Exception:
            log.exception("Unable to save data")
            return None

    log.info("BoB data stored: %s", algo_hash)
    return filepath



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast

from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango
from nbxmpp.modules.dataforms import BooleanField as NBXMPPBooleanField
from nbxmpp.modules.dataforms import DataField as NBXMPPDataField
from nbxmpp.modules.dataforms import DataRecord
from nbxmpp.modules.dataforms import ListField as NBXMPPListField
from nbxmpp.modules.dataforms import ListMultiField as NBXMPPListMultiField
from nbxmpp.modules.dataforms import MultipleDataForm
from nbxmpp.modules.dataforms import SimpleDataForm
from nbxmpp.modules.dataforms import Uri
from nbxmpp.protocol import InvalidJid

from gajim.common import app
from gajim.common.i18n import _
from gajim.common.util.image import get_texture_from_data
from gajim.common.util.text import make_href_markup
from gajim.common.util.text import process_non_spacing_marks

from gajim.gtk.dropdown import GajimDropDown
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import open_uri
from gajim.gtk.widgets import MultiLineLabel

# Options

#  no-scrolling                 No scrollbars
#  form-width                   Minimal form width
#  right-align                  Right align labels
#  hide-fallback-fields         Hide fallback fields in IBR form (ejabberd)
#  left-width                   Width for labels
#  read-only                    Read only mode for fields
#  entry-activates-default      Form entry activates the default widget


class DataFormWidget(Gtk.ScrolledWindow, SignalManager):
    __gsignals__ = {
        "is-valid": (GObject.SignalFlags.RUN_LAST, None, (bool,)),
    }

    def __init__(
        self,
        form_node: SimpleDataForm | MultipleDataForm,
        options: dict[str, Any] | None = None,
    ) -> None:
        Gtk.ScrolledWindow.__init__(
            self, hexpand=True, vexpand=True, overlay_scrolling=False
        )
        SignalManager.__init__(self)

        self.add_css_class("data-form-widget")

        if options is None:
            options = {}

        if options.get("no-scrolling", False):
            self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.NEVER)

        self._form_node = form_node
        if self._form_node.is_reported:
            self._form_widget = DataFormReportedTable(form_node)
        else:
            self._form_widget = FormGrid(form_node, options)
            self._connect(self._form_widget, "form-completed", self._on_form_completed)
        self._original_form_hash = self.get_form_hash()

        self.set_child(self._form_widget)

    def do_unroot(self) -> None:
        Gtk.ScrolledWindow.do_unroot(self)
        self._disconnect_all()
        del self._form_widget
        app.check_finalize(self)

    @property
    def title(self) -> str | None:
        return self._form_widget.title

    @property
    def instructions(self) -> str | None:
        return self._form_widget.instructions

    def validate(self) -> None:
        return self._form_widget.validate(True)

    def get_form(self) -> SimpleDataForm | MultipleDataForm:
        return self._form_node

    def get_submit_form(self) -> SimpleDataForm:
        if self._form_node.is_reported:
            raise ValueError("MultipleDataForm cannot be submitted")
        self._form_node.type_ = "submit"
        return self._form_node

    def get_form_hash(self) -> int:
        return hash(str(self._form_node))

    def was_modified(self) -> bool:
        return self._original_form_hash != self.get_form_hash()

    def reset_form_hash(self) -> None:
        self._original_form_hash = self.get_form_hash()

    def focus_first_entry(self) -> None:
        if self._form_node.is_reported:
            return

        assert isinstance(self._form_widget, FormGrid)
        for row in range(self._form_widget.row_count):
            widget = self._form_widget.get_child_at(1, row)
            if isinstance(widget, Gtk.Entry):
                widget.grab_focus_without_selecting()
                break

    def _on_form_completed(self, *_args: Any) -> None:
        window = cast(Gtk.Window, self.get_root())
        assert window is not None
        window.activate_default()


class FormGrid(Gtk.Grid, SignalManager):
    __gsignals__ = {"form-completed": (GObject.SignalFlags.RUN_LAST, None, ())}

    def __init__(self, form_node: SimpleDataForm, options: dict[str, Any]) -> None:
        Gtk.Grid.__init__(self)
        SignalManager.__init__(self)
        self.set_column_spacing(12)
        self.set_row_spacing(12)
        self.set_halign(Gtk.Align.CENTER)
        self.row_count = 0

        self.rows: list[
            (SizeAdjustment | Title | Instructions | Field | ImageMediaField)
        ] = []

        form_width = options.get("form-width", 435)
        self.set_size_request(form_width, -1)

        self._data_form = form_node

        self.title: str | None = None
        self.instructions: str | None = None

        self._fields = {
            "boolean": BooleanField,
            "fixed": FixedField,
            "list-single": ListSingleField,
            "list-multi": ListMultiField,
            "jid-single": JidSingleField,
            "jid-multi": JidMultiField,
            "text-single": TextSingleField,
            "text-private": TextPrivateField,
            "text-multi": TextMultiField,
        }

        self._add_row(SizeAdjustment(options))

        if form_node.title is not None:
            self.title = form_node.title
            self._add_row(Title(form_node.title))
        if form_node.instructions:
            self.instructions = form_node.instructions
            self._add_row(Instructions(form_node.instructions))

        self._analyse_fields(form_node, options)
        self._parse_form(form_node, options)

    def do_unroot(self) -> None:
        self._disconnect_all()
        for row in self.rows:
            if isinstance(row, Field):
                row.destroy()
        self.rows.clear()
        Gtk.Grid.do_unroot(self)
        app.check_finalize(self)

    def _add_row(
        self, field: SizeAdjustment | Title | Instructions | Field | ImageMediaField
    ) -> None:
        field.add(self, self.row_count)
        self.row_count += 1
        self.rows.append(field)

    @staticmethod
    def _analyse_fields(form_node: SimpleDataForm, options: dict[str, Any]) -> None:
        if "right-align" in options:
            # Don’t overwrite option
            return

        label_lengths = {0}
        for field in form_node.iter_fields():
            if field.type_ == "hidden":
                continue

            if field.label is None:
                continue

            label_lengths.add(len(field.label))

        options["right-align"] = max(label_lengths) < 30

    def _parse_form(self, form_node: SimpleDataForm, options: dict[str, Any]) -> None:
        single_field = (
            len(
                [
                    field
                    for field in form_node.fields
                    if field.type_ not in ("fixed", "hidden")
                ]
            )
            == 1
        )

        for field in form_node.iter_fields():
            if field.type_ == "hidden":
                continue

            if options.get("hide-fallback-fields"):
                if field.var is not None and "fallback" in field.var:
                    continue

            if field.media:
                if not self._add_media_field(field, options):
                    # We don’t understand this media element, ignore it
                    continue

            widget = self._fields[field.type_]
            if single_field and widget is ListSingleField:
                row = widget(field, self, options, treeview=True)
                self._connect(
                    row, "row-activated", self._on_single_list_field_activated
                )
                self._add_row(row)
            else:
                self._add_row(widget(field, self, options))

    def _add_media_field(self, field: NBXMPPDataField, options: dict[str, Any]) -> bool:
        if field.type_ not in ("text-single", "text-private", "text-multi"):
            return False

        assert field.media is not None
        for uri in field.media.uris:
            if not uri.type_.startswith("image/"):
                continue

            if not uri.uri_data.startswith("cid"):
                continue

            self._add_row(ImageMediaField(uri, self, options))
            return True
        return False

    def validate(self, is_valid: bool) -> None:
        value: bool = self._data_form.is_valid() if is_valid else False
        viewport = cast(Gtk.Viewport, self.get_parent())
        dataform_widget = cast(DataFormWidget, viewport.get_parent())
        dataform_widget.emit("is-valid", value)

    def _on_single_list_field_activated(self, widget: ListSingleField) -> None:
        self.emit("form-completed")


class SizeAdjustment:
    def __init__(self, options: dict[str, Any]) -> None:
        self._left_box = Gtk.Box()
        self._right_box = Gtk.Box()

        left_width = options.get("left-width", 100)
        self._left_box.set_size_request(left_width, -1)
        self._right_box.set_hexpand(True)
        self._right_box.set_size_request(250, -1)

    def add(self, form_grid: FormGrid, row_number: int) -> None:
        form_grid.attach(self._left_box, 0, row_number, 1, 1)
        form_grid.attach_next_to(
            self._right_box, self._left_box, Gtk.PositionType.RIGHT, 1, 1
        )


class Title:
    def __init__(self, title: str) -> None:
        self._label = Gtk.Label(
            label=title,
            wrap=True,
            justify=Gtk.Justification.CENTER,
        )
        self._label.add_css_class("title-2")

    def add(self, form_grid: FormGrid, row_number: int) -> None:
        form_grid.attach(self._label, 0, row_number, 2, 1)


class Instructions:
    def __init__(self, instructions: str) -> None:
        self._label = Gtk.Label(wrap=True, justify=Gtk.Justification.CENTER)
        self._label.set_markup(make_href_markup(instructions))

    def add(self, form_grid: FormGrid, row_number: int) -> None:
        form_grid.attach(self._label, 0, row_number, 2, 1)


class Field(GObject.GObject, SignalManager):
    def __init__(
        self, field: NBXMPPDataField, form_grid: FormGrid, options: dict[str, Any]
    ) -> None:
        GObject.GObject.__init__(self)
        SignalManager.__init__(self)

        self._widget: Gtk.Widget | None = None
        self._field = field
        self._form_grid = form_grid
        self._validate_source_id: int | None = None
        self._read_only = options.get("read-only", False)

        self._label = Gtk.Label(
            label=field.label,
            tooltip_text=field.description,
            wrap=True,
            width_chars=15,
            xalign=bool(options.get("right-align")),
        )

        self._warning_image = Gtk.Image.new_from_icon_name(
            "lucide-circle-alert-symbolic"
        )
        self._warning_image.add_css_class("warning")
        self._warning_image.set_visible(False)
        self._warning_image.set_valign(Gtk.Align.CENTER)
        self._warning_image.set_tooltip_text(_("Required"))
        self._warning_box = Gtk.Box()
        self._warning_box.set_size_request(16, -1)
        self._warning_box.append(self._warning_image)

    @property
    def read_only(self) -> bool:
        return self._read_only

    def add(self, form_grid: FormGrid, row_number: int) -> None:
        assert self._widget is not None
        form_grid.attach(self._label, 0, row_number, 1, 1)
        form_grid.attach_next_to(
            self._widget, self._label, Gtk.PositionType.RIGHT, 1, 1
        )
        if self._field.type_ in (
            "jid-single",
            "jid-multi",
            "text-single",
            "text-private",
            "text-multi",
            "list-multi",
        ):
            form_grid.attach_next_to(
                self._warning_box, self._widget, Gtk.PositionType.RIGHT, 1, 1
            )

            is_valid, error = self._field.is_valid()
            self._set_warning(is_valid, error)

    def _set_warning(self, is_valid: bool, error: str | InvalidJid) -> None:
        if not self._field.required and not is_valid and not error:
            # If its not valid and no error is given, its the initial call
            # to show all icons on required fields.
            return

        if error:
            self._warning_image.remove_css_class("warning")
            self._warning_image.add_css_class("error")
        else:
            error = _("Required")
            self._warning_image.remove_css_class("error")
            self._warning_image.add_css_class("warning")
        self._warning_image.set_tooltip_text(str(error))
        self._warning_image.set_visible(not is_valid)

    def _validate(self, delay: bool = True) -> None:
        self._form_grid.validate(False)
        if self._validate_source_id is not None:
            GLib.source_remove(self._validate_source_id)

        def _start_validation() -> None:
            is_valid, error = self._field.is_valid()
            self._set_warning(is_valid, error)
            self._form_grid.validate(is_valid)
            self._validate_source_id = None

        if delay:
            self._validate_source_id = GLib.timeout_add(200, _start_validation)
        else:
            _start_validation()

    def destroy(self) -> None:
        if self._validate_source_id is not None:
            GLib.source_remove(self._validate_source_id)
        self._disconnect_all()
        if self._widget is not None:
            app.check_finalize(self._widget)
        del self._widget
        del self._field
        del self._form_grid


class BooleanField(Field):
    def __init__(
        self, field: NBXMPPBooleanField, form_grid: FormGrid, options: dict[str, Any]
    ) -> None:
        Field.__init__(self, field, form_grid, options)

        if self.read_only:
            label = _("Yes") if field.value else _("No")
            self._widget = Gtk.Label(label=label, selectable=True)
            self._widget.set_xalign(0)
        else:
            self._widget = Gtk.CheckButton()
            self._widget.set_active(field.value)
            self._connect(self._widget, "toggled", self._toggled)
        self._widget.set_valign(Gtk.Align.CENTER)

    def _toggled(self, _widget: Gtk.CheckButton) -> None:
        assert isinstance(self._widget, Gtk.CheckButton)
        assert isinstance(self._field, NBXMPPBooleanField)
        self._field.value = self._widget.get_active()
        self._validate()


class FixedField(Field):
    def __init__(
        self, field: NBXMPPDataField, form_grid: FormGrid, options: dict[str, Any]
    ) -> None:
        Field.__init__(self, field, form_grid, options)

        self._label.set_markup(make_href_markup(field.value))

        # If the value is more than 40 chars it proabably isn’t
        # meant as a section header
        if len(field.value) < 40:
            self._label.add_css_class("field-fixed")
        else:
            self._label.set_xalign(0.5)

    def add(self, form_grid: FormGrid, row_number: int) -> None:
        if len(self._field.value) < 40:
            form_grid.attach(self._label, 0, row_number, 1, 1)
        else:
            form_grid.attach(self._label, 0, row_number, 2, 1)


class ListSingleField(Field):
    __gsignals__ = {"row-activated": (GObject.SignalFlags.RUN_LAST, None, ())}

    def __init__(
        self,
        field: NBXMPPListField,
        form_grid: FormGrid,
        options: dict[str, Any],
        treeview: bool = False,
    ) -> None:
        Field.__init__(self, field, form_grid, options)

        self._unique = treeview
        self._treeview = None
        self._chat_type_drop_down: GajimDropDown[str] | None = None

        if self.read_only:
            self._widget = Gtk.Label(label=field.value, selectable=True)
            self._widget.set_xalign(0)
            return

        if treeview:
            self._setup_treeview(field)
        else:
            self._setup_dropdown(field)

    def destroy(self) -> None:
        Field.destroy(self)
        if self._chat_type_drop_down is not None:
            self._chat_type_drop_down.run_destroy()
        if self._treeview is not None:
            app.check_finalize(self._treeview)
        del self._treeview

    def _setup_dropdown(self, field: NBXMPPListField) -> None:
        data: dict[str, str] = {}
        for value, label in field.iter_options():
            if not label:
                label = value
            data[value] = label

        self._chat_type_drop_down = GajimDropDown(data=data)
        self._chat_type_drop_down.set_valign(Gtk.Align.CENTER)
        self._chat_type_drop_down.select_key(field.value)
        self._widget = self._chat_type_drop_down
        self._connect(self._widget, "notify::selected", self._changed)

    def _changed(self, dropdown: GajimDropDown[str], *args: Any) -> None:
        item = dropdown.get_selected_item()
        assert item is not None
        self._field.value = item.key
        self._validate()

    def _setup_treeview(self, field: NBXMPPListField) -> None:
        self._treeview = ListSingleTreeView(field, self)
        self._connect(self._treeview, "row-activated", self._on_row_activated)
        self._connect(self._treeview, "cursor-changed", self._on_cursor_changed)

        self._widget = Gtk.ScrolledWindow()
        self._widget.set_propagate_natural_height(True)
        self._widget.set_hexpand(True)
        self._widget.set_vexpand(True)
        self._widget.set_min_content_height(100)
        self._widget.set_max_content_height(300)
        self._widget.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self._widget.set_child(self._treeview)

    def _on_cursor_changed(self, *_args: Any) -> None:
        assert self._treeview is not None
        model, treeiter = self._treeview.get_selection().get_selected()
        if treeiter is None:
            return None
        self._field.value = model[treeiter][1]
        self._validate(delay=False)

    def _on_row_activated(self, *_args: Any) -> None:
        self.emit("row-activated")

    def add(self, form_grid: FormGrid, row_number: int) -> None:
        if self._unique:
            assert self._widget is not None
            form_grid.attach(self._widget, 0, row_number, 2, 1)
        else:
            super().add(form_grid, row_number)


class ListSingleTreeView(Gtk.TreeView):
    def __init__(self, field: NBXMPPListField, multi_field: ListSingleField) -> None:
        Gtk.TreeView.__init__(self)

        self._field = field
        self._multi_field = multi_field

        # label, value, tooltip
        self._store = Gtk.ListStore(str, str, str)

        col = Gtk.TreeViewColumn(field.label)
        cell = Gtk.CellRendererText()
        cell.set_property("ellipsize", Pango.EllipsizeMode.END)
        cell.set_property("width-chars", 80)
        col.pack_start(cell, True)
        col.set_attributes(cell, text=0)
        self.append_column(col)

        self.set_headers_visible(True)

        for option in field.options:
            label, value = option
            self._store.append([label, value, label])

        if any(filter(lambda x: len(x[0]) > 80, field.options)):
            self.set_tooltip_column(2)

        self.set_model(self._store)

    def do_unroot(self) -> None:
        Gtk.TreeView.do_unroot(self)
        del self._field
        del self._multi_field
        del self._store
        app.check_finalize(self)


class ListMultiField(Field):
    def __init__(
        self, field: NBXMPPListMultiField, form_grid: FormGrid, options: dict[str, Any]
    ) -> None:
        Field.__init__(self, field, form_grid, options)
        self._label.set_valign(Gtk.Align.START)

        self._treeview = ListMultiTreeView(field, self)

        self._widget = Gtk.ScrolledWindow()
        self._widget.set_propagate_natural_height(True)
        self._widget.set_min_content_height(100)
        self._widget.set_max_content_height(300)
        self._widget.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self._widget.set_child(self._treeview)

    def validate(self) -> None:
        self._validate()

    def destroy(self) -> None:
        Field.destroy(self)
        app.check_finalize(self._treeview)
        del self._treeview


class ListMultiTreeView(Gtk.TreeView, SignalManager):
    def __init__(
        self, field: NBXMPPListMultiField, multi_field: ListMultiField
    ) -> None:
        Gtk.TreeView.__init__(self)
        SignalManager.__init__(self)

        self._field = field
        self._multi_field = multi_field

        # label, value, tooltip, toggled
        self._store = Gtk.ListStore(str, str, str, bool)

        col = Gtk.TreeViewColumn()
        cell = Gtk.CellRendererText()
        cell.set_property("ellipsize", Pango.EllipsizeMode.END)
        cell.set_property("width-chars", 40)
        col.pack_start(cell, True)
        col.set_attributes(cell, text=0)
        self.append_column(col)

        col = Gtk.TreeViewColumn()
        cell = Gtk.CellRendererToggle()
        cell.set_activatable(True)
        cell.set_property("xalign", 1)
        cell.set_property("xpad", 10)
        self._connect(cell, "toggled", self._toggled)
        col.pack_start(cell, True)
        col.set_attributes(cell, active=3)

        if not multi_field.read_only:
            self.append_column(col)

        self.set_headers_visible(False)

        for option in field.options:
            label, value = option
            if multi_field.read_only and value not in field.values:
                continue
            self._store.append([label, value, label, value in field.values])

        if any(filter(lambda x: len(x[0]) > 40, field.options)):
            self.set_tooltip_column(2)

        self.set_model(self._store)

    def do_unroot(self) -> None:
        Gtk.TreeView.do_unroot(self)
        self._disconnect_all()
        del self._field
        del self._multi_field
        del self._store
        app.check_finalize(self)

    def _toggled(self, _renderer: Gtk.CellRendererToggle, path: str) -> None:
        iter_ = self._store.get_iter(path)
        current_value = self._store[iter_][3]
        self._store.set_value(iter_, 3, not current_value)
        self._set_values()
        self._multi_field.validate()

    def _set_values(self) -> None:
        values: list[str] = []
        model = self.get_model()
        assert model is not None
        for row in model:
            if not row[3]:
                continue
            values.append(row[1])
        self._field.values = values


class JidMultiField(Field):
    def __init__(
        self, field: NBXMPPListMultiField, form_grid: FormGrid, options: dict[str, Any]
    ) -> None:
        Field.__init__(self, field, form_grid, options)
        self._label.set_valign(Gtk.Align.START)

        self._treeview = JidMutliTreeView(field, self)

        self._add_button = Gtk.Button(icon_name="lucide-plus-symbolic")
        self._connect(self._add_button, "clicked", self._add_clicked)

        self._remove_button = Gtk.Button(icon_name="lucide-trash-symbolic")
        self._connect(self._remove_button, "clicked", self._remove_clicked)

        self._toolbar = Gtk.Box()
        self._toolbar.add_css_class("toolbar")
        self._toolbar.append(self._add_button)
        self._toolbar.append(self._remove_button)

        self._widget = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self._scrolled_window = Gtk.ScrolledWindow()
        self._scrolled_window.set_propagate_natural_height(True)
        self._scrolled_window.set_min_content_height(100)
        self._scrolled_window.set_max_content_height(300)
        self._scrolled_window.set_child(self._treeview)

        self._widget.append(self._scrolled_window)

        if not self._read_only:
            self._widget.append(self._toolbar)

    def destroy(self) -> None:
        Field.destroy(self)
        app.check_finalize(self._treeview)
        del self._treeview
        del self._scrolled_window
        del self._remove_button
        del self._add_button

    def _add_clicked(self, _widget: Gtk.Button) -> None:
        model = self._treeview.get_model()
        assert isinstance(model, Gtk.ListStore)
        model.append([""])

    def _remove_clicked(self, _widget: Gtk.Button) -> None:
        selected_rows = self._treeview.get_selection().get_selected_rows()
        if selected_rows is None:
            return

        mod, paths = selected_rows
        for path in paths:
            iter_ = mod.get_iter(path)
            model = self._treeview.get_model()
            assert isinstance(model, Gtk.ListStore)
            model.remove(iter_)

        jids: list[str] = []
        model = self._treeview.get_model()
        assert model is not None
        for row in model:
            if not row[0]:
                continue
            jids.append(row[0])

        assert isinstance(self._field, NBXMPPListMultiField)
        self._field.values = jids
        self._validate()

    def validate(self) -> None:
        self._validate()


class JidMutliTreeView(Gtk.TreeView, SignalManager):
    def __init__(self, field: NBXMPPListMultiField, multi_field: JidMultiField) -> None:
        Gtk.TreeView.__init__(self)
        SignalManager.__init__(self)

        self._field = field
        self._multi_field = multi_field

        self._store = Gtk.ListStore(str)

        col = Gtk.TreeViewColumn()
        cell = Gtk.CellRendererText()
        cell.set_property("editable", True)
        cell.set_property("placeholder-text", "user@example.org")
        self._connect(cell, "edited", self._jid_edited)
        col.pack_start(cell, True)
        col.set_attributes(cell, text=0)
        self.append_column(col)

        self.set_headers_visible(False)

        for value in field.values:
            self._store.append([value])

        self.set_model(self._store)

    def do_unroot(self) -> None:
        Gtk.TreeView.do_unroot(self)
        self._disconnect_all()
        del self._field
        del self._multi_field
        del self._store
        app.check_finalize(self)

    def _jid_edited(
        self, _renderer: Gtk.CellRendererText, path: str, new_text: str
    ) -> None:
        iter_ = self._store.get_iter(path)
        self._store.set_value(iter_, 0, new_text)
        self._set_values()
        self._multi_field.validate()

    def _set_values(self) -> None:
        jids: list[str] = []
        for row in self._store:
            if not row[0]:
                continue
            jids.append(row[0])
        self._field.values = jids


class TextSingleField(Field):
    def __init__(
        self, field: NBXMPPDataField, form_grid: FormGrid, options: dict[str, Any]
    ) -> None:
        Field.__init__(self, field, form_grid, options)

        if self.read_only:
            self._widget = Gtk.Label(label=field.value, selectable=True)
            self._widget.set_xalign(0)
            self._widget.set_selectable(True)
        else:
            self._widget = Gtk.Entry()
            self._widget.set_text(field.value)
            self._connect(self._widget, "changed", self._changed)
            if options.get("entry-activates-default", False):
                self._widget.set_activates_default(True)
        self._widget.set_valign(Gtk.Align.CENTER)

    def _changed(self, _widget: Gtk.Entry) -> None:
        assert isinstance(self._widget, Gtk.Entry)
        self._field.value = self._widget.get_text()
        self._validate()


class TextPrivateField(TextSingleField):
    def __init__(
        self, field: NBXMPPDataField, form_grid: FormGrid, options: dict[str, Any]
    ) -> None:
        TextSingleField.__init__(self, field, form_grid, options)
        if isinstance(self._widget, Gtk.Entry):
            self._widget.set_input_purpose(Gtk.InputPurpose.PASSWORD)
            self._widget.set_visibility(False)


class JidSingleField(TextSingleField):
    def __init__(
        self, field: NBXMPPDataField, form_grid: FormGrid, options: dict[str, Any]
    ) -> None:
        TextSingleField.__init__(self, field, form_grid, options)


class TextMultiField(Field):
    def __init__(
        self, field: NBXMPPDataField, form_grid: FormGrid, options: dict[str, Any]
    ) -> None:
        Field.__init__(self, field, form_grid, options)
        self._label.set_valign(Gtk.Align.START)

        length = len(field.value)

        if self.read_only:
            self._textview = MultiLineLabel(
                label=process_non_spacing_marks(field.value), xalign=0
            )

        else:
            self._textview = Gtk.TextView()
            self._textview.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)
            self._textview.get_buffer().set_text(field.value)
            self._connect(self._textview.get_buffer(), "changed", self._changed)

        if self.read_only and length < 500:
            self._widget = self._textview
        else:
            self._widget = Gtk.ScrolledWindow()
            self._widget.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
            self._widget.set_propagate_natural_height(True)
            self._widget.set_min_content_height(100)

            self._widget.set_child(self._textview)

    def _changed(self, widget: Gtk.TextBuffer) -> None:
        self._field.value = widget.get_text(*widget.get_bounds(), False)
        self._validate()

    def destroy(self) -> None:
        Field.destroy(self)
        app.check_finalize(self._textview)
        del self._textview


class ImageMediaField:
    def __init__(self, uri: Uri, form_grid: FormGrid, _options: dict[str, Any]) -> None:
        self._uri = uri
        self._form_grid = form_grid

        filename = uri.uri_data.split(":")[1].split("@")[0]
        data = app.bob_cache.get(filename)
        if data is None:
            self._image = Gtk.Image()
            return

        texture = get_texture_from_data(data, 170)
        self._image = Gtk.Image.new_from_paintable(texture)
        self._image.set_halign(Gtk.Align.CENTER)
        self._image.set_pixel_size(170)
        self._image.add_css_class("preview-image")

    def add(self, form_grid: FormGrid, row_number: int) -> None:
        form_grid.attach(self._image, 1, row_number, 1, 1)


class FakeDataFormWidget(Gtk.ScrolledWindow, SignalManager):
    def __init__(self, fields: dict[str, str]) -> None:
        Gtk.ScrolledWindow.__init__(self)
        SignalManager.__init__(self)

        self.set_hexpand(True)
        self.set_vexpand(True)
        self.set_overlay_scrolling(False)
        self.add_css_class("data-form-widget")

        self._grid = Gtk.Grid()
        self._grid.set_column_spacing(12)
        self._grid.set_row_spacing(12)
        self._grid.set_halign(Gtk.Align.CENTER)

        self._fields = fields
        self._entries: dict[str, Gtk.Entry] = {}
        self._row_count = 0

        instructions = fields.pop("instructions", None)
        if instructions is not None:
            label = Gtk.Label(
                label=instructions,
                justify=Gtk.Justification.CENTER,
                max_width_chars=40,
                wrap=True,
            )

            self._grid.attach(label, 0, self._row_count, 2, 1)
            self._row_count += 1

        redirect_url = fields.pop("redirect-url", None)
        if not fields and redirect_url is not None:
            # Server wants to redirect registration
            button = Gtk.Button(label="Register")
            button.set_halign(Gtk.Align.CENTER)
            button.add_css_class("suggested-action")
            self._connect(button, "clicked", lambda *args: open_uri(redirect_url))  # type: ignore
            self._grid.attach(button, 0, self._row_count, 2, 1)
        else:
            self._add_fields()
        self.set_child(self._grid)

    def do_unroot(self) -> None:
        Gtk.ScrolledWindow.do_unroot(self)
        self._disconnect_all()
        app.check_finalize(self._grid)
        del self._grid
        app.check_finalize(self)

    def _add_fields(self) -> None:
        for name, value in self._fields.items():
            if name in ("key", "x", "registered"):
                continue

            label = Gtk.Label(label=name.capitalize())
            label.set_xalign(1)
            self._grid.attach(label, 0, self._row_count, 1, 1)
            self._row_count += 1

            entry = Gtk.Entry()
            entry.set_text(value)
            if name == "password":
                entry.set_input_purpose(Gtk.InputPurpose.PASSWORD)
                entry.set_visibility(False)
            self._entries[name] = entry

            self._grid.attach_next_to(entry, label, Gtk.PositionType.RIGHT, 1, 1)

    def get_submit_form(self) -> dict[str, str]:
        fields: dict[str, str] = {}
        for name, entry in self._entries.items():
            fields[name] = entry.get_text()
        return fields


class DataFormReportedTable(Gtk.Grid, SignalManager):
    """
    A widget representing the results of a command as a table, when the data
    form contains a <reported> element.
    """

    def __init__(self, form_node: MultipleDataForm) -> None:
        Gtk.Grid.__init__(self, row_spacing=10)
        SignalManager.__init__(self)
        self._form_node = form_node

        self.title = form_node.title
        self.instructions = form_node.instructions

        self._process_form()
        self._build_list_store()
        self._build_treeview()
        self._build_scrolled_window()

    def do_unroot(self) -> None:
        Gtk.Grid.do_unroot(self)
        self._disconnect_all()
        del self._list_store
        del self._treeview
        app.check_finalize(self)

    def _process_form(self) -> None:
        self._column_names: list[str] = []
        self._column_types: list[str] = []
        for field in self._form_node.reported.iter_fields():
            self._column_types.append(field.type_)
            self._column_names.append(field.label)

        rows: list[list[str]] = [
            [f.value for f in record.fields]
            for record in cast(list[DataRecord], self._form_node.items)
        ]

        self._rows = rows

    def _build_list_store(self) -> None:
        self._list_store = Gtk.ListStore(*(str for _ in self._column_names))
        for row in self._rows:
            self._list_store.append(row)

    def _build_treeview(self) -> None:
        self._treeview = Gtk.TreeView(model=self._list_store.filter_new())
        self._connect(self._treeview, "row-activated", self._on_row_activate)

        for i, column_title in enumerate(self._column_names):
            renderer = Gtk.CellRendererText()
            column = Gtk.TreeViewColumn(column_title, renderer, text=i)
            self._treeview.append_column(column)

    def _build_scrolled_window(self) -> None:
        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_vexpand(True)
        scrolled_window.set_hexpand(True)
        scrolled_window.set_child(self._treeview)

        title = f"<big>{GLib.markup_escape_text(self._form_node.title)}</big>"
        self.attach(Gtk.Label(label=title, use_markup=True), 0, 0, 1, 1)
        self.attach(scrolled_window, 0, 2, 1, 1)

    def _on_row_activate(
        self,
        _tree_view: Gtk.TreeView,
        path: Gtk.TreePath,
        column: Gtk.TreeViewColumn,
    ) -> None:
        row = self._rows[int(path.to_string())]
        col = self._column_names.index(column.get_title())
        # If the activated cell is a JID, start a chat with it.
        if self._column_types[col] == "jid-single":
            dest: str | None = row[col]
        else:
            # else, try to find a column that is a JID
            for i, t in enumerate(self._column_types):
                if t == "jid-single":
                    dest = row[i]
                    break
            else:
                dest = None
        if dest:
            app.app.activate_action("start-chat", GLib.Variant("as", [dest, ""]))

    def validate(self, is_valid: bool) -> None:
        viewport = cast(Gtk.Viewport, self.get_parent())
        dataform_widget = cast(DataFormWidget, viewport.get_parent())
        dataform_widget.emit("is-valid", is_valid)



# Copyright (C) 2009 Bruno Tarquini <btarquini AT gmail.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

import logging
import os
import sys
import time
from collections.abc import Callable
from datetime import datetime

from gajim.common import app
from gajim.common import configpaths
from gajim.common.i18n import _

LogCallback = Callable[[str], None]


def parseLogLevel(arg: str) -> int:
    """
    Either numeric value or level name from logging module
    """
    if arg.isdigit():
        return int(arg)
    if arg.isupper() and hasattr(logging, arg):
        return getattr(logging, arg)
    print(_("%s is not a valid loglevel") % repr(arg), file=sys.stderr)
    return 0


def parseLogTarget(arg: str) -> str:
    """
    [gajim.]c.x.y  ->  gajim.c.x.y
    .other_logger  ->  other_logger
    <None>         ->  gajim
    """
    arg = arg.lower()
    if not arg:
        return "gajim"
    if arg.startswith("."):
        return arg[1:]
    if arg.startswith("gajim"):
        return arg
    return "gajim." + arg


def parseAndSetLogLevels(arg: str) -> None:
    """
    [=]LOGLEVEL     ->  gajim=LOGLEVEL
    gajim=LOGLEVEL  ->  gajim=LOGLEVEL
    .other=10       ->  other=10
    .=10            ->  <nothing>
    c.x.y=c.z=20    ->  gajim.c.x.y=20
                        gajim.c.z=20
    gajim=10,c.x=20 ->  gajim=10
                        gajim.c.x=20
    """
    for directive in arg.split(","):
        directive = directive.strip()
        if not directive:
            continue
        if "=" not in directive:
            directive = "=" + directive
        targets, level = directive.rsplit("=", 1)
        level = parseLogLevel(level.strip())
        for target in targets.split("="):
            target = parseLogTarget(target.strip())
            if target:
                logging.getLogger(target).setLevel(level)
                print("Logger %s level set to %d" % (target, level), file=sys.stderr)


class Colors:
    NONE = chr(27) + "[0m"
    BLACK = chr(27) + "[30m"
    RED = chr(27) + "[31m"
    GREEN = chr(27) + "[32m"
    BROWN = chr(27) + "[33m"
    BLUE = chr(27) + "[34m"
    MAGENTA = chr(27) + "[35m"
    CYAN = chr(27) + "[36m"
    LIGHT_GRAY = chr(27) + "[37m"
    DARK_GRAY = chr(27) + "[30;1m"
    BRIGHT_RED = chr(27) + "[31;1m"
    BRIGHT_GREEN = chr(27) + "[32;1m"
    YELLOW = chr(27) + "[33;1m"
    BRIGHT_BLUE = chr(27) + "[34;1m"
    PURPLE = chr(27) + "[35;1m"
    BRIGHT_CYAN = chr(27) + "[36;1m"
    WHITE = chr(27) + "[37;1m"


def colorize(text: str, color: str) -> str:
    return color + text + Colors.NONE


class LogConsoleHandler(logging.StreamHandler):  # type: ignore
    def __init__(self) -> None:
        super().__init__()  # type: ignore
        self._callback: LogCallback | None = None

    def emit(self, record: logging.LogRecord) -> None:
        if record.levelno < logging.WARNING:
            return

        msg = self.format(record)
        app.logging_records.append(msg)
        if self._callback is not None:
            self._callback(msg)

    def set_callback(self, func: LogCallback | None) -> None:
        self._callback = func


logging.Formatter.default_msec_format = None


class FancyFormatter(logging.Formatter):
    """
    An eye-candy formatter with Colors
    """

    colors_mapping = {
        "DEBUG": Colors.BLUE,
        "INFO": Colors.GREEN,
        "WARNING": Colors.BROWN,
        "ERROR": Colors.RED,
        "CRITICAL": Colors.BRIGHT_RED,
    }

    def __init__(self, fmt: str | None = None, use_color: bool = False) -> None:
        logging.Formatter.__init__(self, fmt)
        self.use_color = use_color

    def format(self, record: logging.LogRecord) -> str:
        level = record.levelname
        record.levelname = "(%s)" % level[0]

        if self.use_color:
            c = FancyFormatter.colors_mapping.get(level, "")
            record.levelname = colorize(record.levelname, c)
            record.name = "%-25s" % colorize(record.name, Colors.CYAN)
        else:
            record.name = "%-25s|" % record.name

        return logging.Formatter.format(self, record)


def init() -> None:
    """
    Initialize the logging system
    """

    if app.get_debug_mode():
        _cleanup_debug_logs()
        _redirect_output()

    use_color = False
    if os.name != "nt":
        use_color = sys.stderr.isatty()

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(
        FancyFormatter("%(asctime)s %(levelname)s %(name)-35s %(message)s", use_color)
    )

    root_log = logging.getLogger("gajim")
    root_log.setLevel(logging.WARNING)
    root_log.addHandler(_log_console_handler)
    root_log.addHandler(stream_handler)
    root_log.propagate = False

    root_log = logging.getLogger("nbxmpp")
    root_log.setLevel(logging.WARNING)
    root_log.addHandler(_log_console_handler)
    root_log.addHandler(stream_handler)
    root_log.propagate = False

    root_log = logging.getLogger("gnupg")
    root_log.setLevel(logging.WARNING)
    root_log.addHandler(_log_console_handler)
    root_log.addHandler(stream_handler)
    root_log.propagate = False

    root_log = logging.getLogger("omemo_dr")
    root_log.setLevel(logging.WARNING)
    root_log.addHandler(_log_console_handler)
    root_log.addHandler(stream_handler)
    root_log.propagate = False

    if os.environ.get("GAJIM_LEAK"):
        logging.getLogger("gajim.leak").setLevel(logging.DEBUG)

    if os.environ.get("GAJIM_DEBUG_SQL"):
        logging.getLogger("gajim.c.storage").setLevel(logging.DEBUG)

    # GAJIM_DEBUG is set only on Windows when using Gajim-Debug.exe
    # Gajim-Debug.exe shows a command line prompt and we want to redirect
    # log output to it
    if app.get_debug_mode() or os.environ.get("GAJIM_DEBUG", False):  # noqa: PLW1508
        set_verbose()


def set_loglevels(loglevels_string: str) -> None:
    parseAndSetLogLevels(loglevels_string)


def set_verbose() -> None:
    parseAndSetLogLevels("gajim=DEBUG")
    parseAndSetLogLevels(".nbxmpp=INFO")
    parseAndSetLogLevels(".omemo_dr=DEBUG")


def set_quiet() -> None:
    parseAndSetLogLevels("gajim=CRITICAL")
    parseAndSetLogLevels(".nbxmpp=CRITICAL")
    parseAndSetLogLevels(".omemo_dr=CRITICAL")


def get_log_console_handler() -> LogConsoleHandler:
    return _log_console_handler


def _redirect_output() -> None:
    debug_folder = configpaths.get("DEBUG")
    date = datetime.today().strftime("%Y-%m-%d-%H%M%S")
    filename = "%s-debug.log" % date
    fd = open(debug_folder / filename, "a", encoding="utf8")
    sys.stderr = sys.stdout = fd


def _cleanup_debug_logs() -> None:
    debug_folder = configpaths.get("DEBUG")
    debug_files = list(debug_folder.glob("*-debug.log*"))
    now = time.time()
    for file in debug_files:
        # Delete everything older than 3 days
        if file.stat().st_ctime < now - 259200:
            file.unlink()


_log_console_handler = LogConsoleHandler()
_log_console_handler.setFormatter(
    logging.Formatter(
        "%(asctime)s (%(levelname).1s) %(name)-35s | %(message)s\n",
    )
)


if __name__ == "__main__":
    init()

    set_loglevels("gajim.c=DEBUG,INFO")

    log = logging.getLogger("gajim")
    log.debug("debug")
    log.info("info")
    log.warning("warn")
    log.error("error")
    log.critical("critical")

    log = logging.getLogger("gajim.c.x.dispatcher")
    log.debug("debug")
    log.info("info")
    log.warning("warn")
    log.error("error")
    log.critical("critical")



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import sys
import unicodedata
from collections.abc import Callable
from pathlib import Path

from gi.repository import Gio
from gi.repository import GLib
from nbxmpp.protocol import JID

WIN_PATH_BLACKLIST = ["\\", "/", ":", "*", "?", "？", '"', "<", ">", "|", "\0"]
WIN_RESERVED_FILENAMES = [
    "CON",
    "PRN",
    "AUX",
    "NUL",
    "COM1",
    "COM2",
    "COM3",
    "COM4",
    "COM5",
    "COM6",
    "COM7",
    "COM8",
    "COM9",
    "LPT1",
    "LPT2",
    "LPT3",
    "LPT4",
    "LPT5",
    "LPT6",
    "LPT7",
    "LPT8",
    "LPT9",
]


def sanitize_filename(filename: str, max_length: int = 50) -> str:
    """
    Sanitize filename of:
     - characters used to obfuscate file names/extensions
     - elements not allowed on Windows
       https://docs.microsoft.com/en-us/windows/win32/fileio/naming-a-file
     - shorten to max length
    """

    # Remove right-to-left override U+202E (commonly used to spoof extensions)
    filename = filename.replace("\u202e", "")

    if sys.platform == "win32":
        filename = sanitize_windows_filename(filename)

    extension = Path(filename).suffix[:10]
    filename = Path(filename).stem

    if max_length > 0:
        final_length = max_length - len(extension)
        filename = filename[:final_length]

    return f"{filename}{extension}"


def sanitize_windows_filename(filename: str) -> str:
    filename = "".join(char for char in filename if char not in WIN_PATH_BLACKLIST)
    filename = "".join(char for char in filename if ord(char) > 31)

    filename = unicodedata.normalize("NFKD", filename)
    filename = filename.rstrip(". ")
    filename = filename.strip()

    if all(char == "." for char in filename):
        filename = f"__{filename}"
    if filename.upper() in WIN_RESERVED_FILENAMES:
        filename = f"__{filename}"
    if len(filename) == 0:
        filename = "__"

    return filename


def make_path_from_jid(base_path: Path, jid: JID) -> Path:
    assert jid.domain is not None
    domain = jid.domain[:50]

    if jid.localpart is None:
        return base_path / domain

    path = base_path / domain / sanitize_filename(jid.localpart)
    if jid.resource is not None:
        return path / sanitize_filename(jid.resource, max_length=30)
    return path


def write_file_async(
    path: Path,
    data: bytes,
    callback: Callable[[bool, GLib.Error | None, Any], Any],
    user_data: Any | None = None,
):

    def _on_write_finished(
        outputstream: Gio.OutputStream, result: Gio.AsyncResult, _data: bytes
    ) -> None:
        try:
            successful, _bytes_written = outputstream.write_all_finish(result)
        except GLib.Error as error:
            callback(False, error, user_data)
        else:
            callback(successful, None, user_data)

    def _on_file_created(file: Gio.File, result: Gio.AsyncResult) -> None:
        try:
            outputstream = file.create_finish(result)
        except GLib.Error as error:
            callback(False, error, user_data)
            return

        # Pass data as user_data to the callback, because
        # write_all_async() takes no reference to the data
        # and python gc collects it before the data is written
        outputstream.write_all_async(
            data, GLib.PRIORITY_DEFAULT, None, _on_write_finished, data
        )

    file = Gio.File.new_for_path(str(path))
    file.create_async(
        Gio.FileCreateFlags.PRIVATE, GLib.PRIORITY_DEFAULT, None, _on_file_created
    )


def load_file_async(
    path: Path,
    callback: Callable[[bytes | None, GLib.Error | None, Any], Any],
    user_data: Any | None = None,
) -> None:

    def _on_load_finished(file: Gio.File, result: Gio.AsyncResult) -> None:

        try:
            _, contents, _ = file.load_contents_finish(result)
        except GLib.Error as error:
            callback(None, error, user_data)
        else:
            # "contents" may be an empty bytes object
            callback(contents or None, None, user_data)

    file = Gio.File.new_for_path(str(path))
    file.load_contents_async(None, _on_load_finished)



# Copyright (C) 2009 Stephan Erb <steve-e AT h3c.de>
# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0115: Entity Capabilities

from __future__ import annotations

from typing import Any
from typing import cast

import weakref
from collections import defaultdict
from collections.abc import Callable

from nbxmpp.errors import StanzaError
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.protocol import Presence
from nbxmpp.structs import DiscoIdentity
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import EntityCapsData
from nbxmpp.structs import PresenceProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import Task as nbxmpp_Task
from nbxmpp.util import compute_caps_hash

from gajim.common import app
from gajim.common import types
from gajim.common.const import COMMON_FEATURES
from gajim.common.const import Entity
from gajim.common.helpers import get_optional_features
from gajim.common.modules.base import BaseModule
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.task_manager import Task


class Caps(BaseModule):
    _nbxmpp_extends = "EntityCaps"
    _nbxmpp_methods = ["set_caps"]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.handlers = [
            StanzaHandler(
                name="presence",
                callback=self._entity_caps,
                typ="available",
                ns=Namespace.CAPS,
                priority=51,
            ),
        ]

        self._identities = [DiscoIdentity(category="client", type="pc", name="Gajim")]

        self._queued_tasks_by_hash: defaultdict[str, set[EntityCapsTask]] = defaultdict(
            set
        )
        self._queued_tasks_by_jid: dict[JID, EntityCapsTask] = {}

    def get_own_caps(self) -> EntityCapsData | None:
        return self._nbxmpp("EntityCaps").caps

    def _queue_task(self, task: EntityCapsTask) -> None:
        old_task = self._get_task(task.entity.jid)
        if old_task is not None:
            self._remove_task(old_task)

        self._log.info("Queue query for hash %s", task.entity.hash)
        self._queued_tasks_by_hash[task.entity.hash].add(task)
        self._queued_tasks_by_jid[task.entity.jid] = task
        app.task_manager.add_task(task)

    def _get_task(self, jid: JID) -> EntityCapsTask | None:
        return self._queued_tasks_by_jid.get(jid)

    def _get_similar_tasks(self, task: EntityCapsTask) -> set[EntityCapsTask]:
        return self._queued_tasks_by_hash.pop(task.entity.hash)

    def _remove_task(self, task: EntityCapsTask) -> None:
        task.set_obsolete()
        del self._queued_tasks_by_jid[task.entity.jid]
        self._queued_tasks_by_hash[task.entity.hash].discard(task)

    def _remove_all_tasks(self) -> None:
        for task in self._queued_tasks_by_jid.values():
            task.set_obsolete()
        self._queued_tasks_by_jid.clear()
        self._queued_tasks_by_hash.clear()

    def _entity_caps(
        self,
        _con: types.NBXMPPClient,
        _stanza: Presence,
        properties: PresenceProperties,
    ) -> None:

        if properties.is_self_presence:
            return

        if properties.entity_caps is None:
            return

        if properties.muc_user is not None:
            # Don’t query MUC participants
            return

        task = EntityCapsTask(self._account, properties, self._execute_task)

        self._log.info("Received %s", task.entity)

        disco_info = app.storage.cache.get_caps_entry(
            task.entity.method, task.entity.hash
        )
        if disco_info is None:
            self._queue_task(task)
            return

        assert properties.jid is not None
        app.storage.cache.set_last_disco_info(
            properties.jid, disco_info, cache_only=True
        )

        contact = self._con.get_module("Contacts").get_contact(properties.jid)
        contact.notify("caps-update")

    def _execute_task(self, task: EntityCapsTask) -> None:
        self._log.info("Request %s from %s", task.entity.hash, task.entity.jid)
        self._con.get_module("Discovery").disco_info(
            task.entity.jid,
            node=f"{task.entity.node}#{task.entity.hash}",
            callback=self._on_disco_info,
            user_data=task.entity.jid,
        )

    def _on_disco_info(self, nbxmpp_task: nbxmpp_Task) -> None:
        jid = nbxmpp_task.get_user_data()
        task = self._get_task(jid)
        if task is None:
            self._log.info("Task not found for %s", jid)
            return

        self._remove_task(task)

        try:
            disco_info = cast(DiscoInfo, nbxmpp_task.finish())
        except StanzaError as error:
            self._log.warning(error)
            return

        self._log.info("Disco Info received: %s", disco_info.jid)

        try:
            compute_caps_hash(disco_info)
        except Exception as error:
            self._log.warning("Disco info malformed: %s %s", disco_info.jid, error)
            return

        assert disco_info.jid is not None
        caps_hash = disco_info.get_caps_hash()
        assert caps_hash is not None
        app.storage.cache.add_caps_entry(
            disco_info.jid, task.entity.method, caps_hash, disco_info
        )

        self._log.info("Finished query for %s", task.entity.hash)

        tasks = self._get_similar_tasks(task)

        for task in tasks:
            self._remove_task(task)
            self._log.info("Update %s", task.entity.jid)
            contact = self._con.get_module("Contacts").get_contact(task.entity.jid)
            app.storage.cache.set_last_disco_info(
                task.entity.jid, disco_info, cache_only=True
            )
            contact.notify("caps-update")

    def update_caps(self) -> None:
        if not app.account_is_connected(self._account):
            return

        optional_features = get_optional_features(self._account)
        self.set_caps(
            self._identities, COMMON_FEATURES + optional_features, "https://gajim.org"
        )

        if not app.account_is_available(self._account):
            return

        client = app.get_client(self._account)
        client.change_status(client.status, client.status_message)

    def cleanup(self) -> None:
        BaseModule.cleanup(self)
        self._remove_all_tasks()


class EntityCapsTask(Task):
    def __init__(
        self, account: str, properties: PresenceProperties, callback: Callable[..., Any]
    ) -> None:
        Task.__init__(self)
        self._account = account
        self._callback = weakref.WeakMethod(callback)

        assert properties.jid is not None
        assert properties.entity_caps is not None
        self.entity = Entity(
            jid=properties.jid,
            node=properties.entity_caps.node,
            hash=properties.entity_caps.ver,
            method=properties.entity_caps.hash,
        )

        self._from_muc = properties.from_muc

    def execute(self) -> None:
        callback = self._callback()
        if callback is not None:
            callback(self)

    def preconditions_met(self) -> bool:
        try:
            client = app.get_client(self._account)
        except Exception:
            return False

        if self._from_muc:
            contact = client.get_module("Contacts").get_contact(
                self.entity.jid.bare, groupchat=True
            )

            assert isinstance(contact, GroupchatContact)
            if not contact.is_joined:
                self.set_obsolete()
                return False

        return client.is_available()

    def __repr__(self) -> str:
        return f"Entity Caps ({self.entity.jid} {self.entity.hash})"

    def __hash__(self) -> int:
        return hash(self.entity)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import cast
from typing import Literal
from typing import overload

import logging

from gi.repository import Gtk
from nbxmpp.errors import ChangePasswordStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.dataforms import SimpleDataForm
from nbxmpp.task import Task

from gajim.common import app
from gajim.common import passwords
from gajim.common.helpers import to_user_string
from gajim.common.i18n import _

from gajim.gtk.assistant import Assistant
from gajim.gtk.assistant import AssistantErrorPage
from gajim.gtk.assistant import AssistantPage
from gajim.gtk.assistant import AssistantSuccessPage
from gajim.gtk.dataform import DataFormWidget
from gajim.gtk.util.misc import ensure_not_destroyed

log = logging.getLogger("gajim.gtk.change_password")


class ChangePassword(Assistant):
    def __init__(self, account: str) -> None:
        Assistant.__init__(self)

        self.account = account
        self._client = app.get_client(account)
        self._destroyed = False

        cur_password = passwords.get_password(self.account)
        assert cur_password is not None

        self.add_button("apply", _("Change"), "suggested-action", complete=True)
        self.add_button("close", _("Close"))
        self.add_button("back", _("Back"))

        self.add_pages(
            {
                "password": EnterPassword(cur_password),
                "next_stage": NextStage(),
                "error": Error(),
                "success": Success(),
            }
        )

        progress = self.add_default_page("progress")
        progress.set_title(_("Changing Password..."))
        progress.set_text(_("Trying to change password..."))

        self._connect(self, "button-clicked", self._on_button_clicked)

    @overload
    def get_page(self, name: Literal["password"]) -> EnterPassword: ...

    @overload
    def get_page(self, name: Literal["next_stage"]) -> NextStage: ...

    @overload
    def get_page(self, name: Literal["error"]) -> Error: ...

    @overload
    def get_page(self, name: Literal["success"]) -> Success: ...

    def get_page(self, name: str) -> AssistantPage:
        return self._pages[name]

    def _on_button_clicked(self, _assistant: Assistant, button_name: str) -> None:
        page = self.get_current_page()
        if button_name == "apply":
            self.show_page("progress", Gtk.StackTransitionType.SLIDE_LEFT)
            self._on_apply(next_stage=page == "next_stage")

        elif button_name == "back":
            self.show_page("password", Gtk.StackTransitionType.SLIDE_RIGHT)

        elif button_name == "close":
            self.close()

    def _on_apply(self, next_stage: bool = False) -> None:
        if next_stage:
            # TODO: Does not apply sasl prep profile
            form = self.get_page("next_stage").get_submit_form()
            self._client.get_module("Register").change_password_with_form(
                form, callback=self._on_change_password
            )
        else:
            password = self.get_page("password").get_password()
            self._client.get_module("Register").change_password(
                password, callback=self._on_change_password
            )

    @ensure_not_destroyed
    def _on_change_password(self, task: Task) -> None:
        try:
            task.finish()
        except ChangePasswordStanzaError as error:
            form = cast(SimpleDataForm, error.get_form())
            self.get_page("next_stage").set_form(form)
            self.show_page("next_stage", Gtk.StackTransitionType.SLIDE_LEFT)

        except StanzaError as error:
            error_text = to_user_string(error)
            self.get_page("error").set_text(error_text)
            self.show_page("error", Gtk.StackTransitionType.SLIDE_LEFT)

        else:
            password = self.get_page("password").get_password()
            passwords.save_password(self.account, password)
            self.show_page("success")

    def _cleanup(self) -> None:
        Assistant._cleanup(self)
        self._destroyed = True


class EnterPassword(AssistantPage):
    def __init__(self, cur_password: str) -> None:
        AssistantPage.__init__(self)
        self.complete = False
        self.title = _("Change Password")
        self._cur_password = cur_password

        heading = Gtk.Label(
            label=_("Change Password"),
            wrap=True,
            max_width_chars=30,
            halign=Gtk.Align.CENTER,
            justify=Gtk.Justification.CENTER,
        )

        label = Gtk.Label(
            label=_("Please enter your new password."),
            wrap=True,
            max_width_chars=50,
            halign=Gtk.Align.CENTER,
            justify=Gtk.Justification.CENTER,
            margin_bottom=12,
        )

        self._cur_password_entry = Gtk.PasswordEntry(
            show_peek_icon=True,
            placeholder_text=_("Enter current password..."),
            valign=Gtk.Align.END,
        )
        self._connect(self._cur_password_entry, "changed", self._on_changed)

        self._password1_entry = Gtk.PasswordEntry(
            show_peek_icon=True,
            placeholder_text=_("Enter new password..."),
            valign=Gtk.Align.END,
        )
        self._connect(self._password1_entry, "changed", self._on_changed)

        self._password2_entry = Gtk.PasswordEntry(
            activates_default=True,
            show_peek_icon=True,
            placeholder_text=_("Confirm new password..."),
            valign=Gtk.Align.START,
        )
        self._connect(self._password2_entry, "changed", self._on_changed)

        box = Gtk.Box(spacing=12, halign=Gtk.Align.CENTER)
        self._warning_label = Gtk.Label(max_width_chars=50)
        self._icon = Gtk.Image(icon_name="lucide-circle-alert-symbolic", visible=False)
        box.append(self._icon)
        box.append(self._warning_label)

        self.append(heading)
        self.append(label)
        self.append(self._cur_password_entry)
        self.append(self._password1_entry)
        self.append(self._password2_entry)
        self.append(box)
        self._hide_warning()

    def _hide_warning(self) -> None:
        self._icon.set_visible(False)
        self._warning_label.set_text("")

    def _show_warning(self, text: str) -> None:
        self._icon.set_visible(True)
        self._warning_label.set_text(text)

    def _on_changed(self, _entry: Gtk.Entry) -> None:
        cur_password = self._cur_password_entry.get_text()
        if cur_password != self._cur_password:
            self._show_warning(_("Current password incorrect"))
            self._set_complete(False)
            return

        password1 = self._password1_entry.get_text()
        if not password1:
            self._show_warning(_("Passwords do not match"))
            self._set_complete(False)
            return

        password2 = self._password2_entry.get_text()
        if password1 != password2:
            self._show_warning(_("Passwords do not match"))
            self._set_complete(False)
            return

        self._hide_warning()
        self._set_complete(True)

    def _set_complete(self, state: bool) -> None:
        self.complete = state
        self.update_page_complete()

    def get_password(self) -> str:
        return self._password1_entry.get_text()

    def get_visible_buttons(self) -> list[str]:
        return ["apply"]


class NextStage(AssistantPage):
    def __init__(self) -> None:
        AssistantPage.__init__(self)
        self.set_valign(Gtk.Align.FILL)
        self.complete = False
        self.title = _("Change Password")
        self._current_form = None

    def _on_is_valid(self, _widget: Gtk.Widget, is_valid: bool) -> None:
        self.complete = is_valid
        self.update_page_complete()

    def set_form(self, form: SimpleDataForm) -> None:
        if self._current_form is not None:
            self.remove(self._current_form)
        self._current_form = DataFormWidget(form)
        self._connect(self._current_form, "is-valid", self._on_is_valid)
        self._current_form.validate()
        self.append(self._current_form)

    def get_submit_form(self) -> SimpleDataForm:
        assert self._current_form is not None
        form = self._current_form.get_submit_form()
        assert isinstance(form, SimpleDataForm)
        return form

    def get_visible_buttons(self) -> list[str]:
        return ["apply"]


class Error(AssistantErrorPage):
    def __init__(self) -> None:
        AssistantErrorPage.__init__(self)
        self.set_title(_("Password Change Failed"))
        self.set_heading(_("Password Change Failed"))
        self.set_text(_("An error occurred while trying to change your password."))

    def get_visible_buttons(self) -> list[str]:
        return ["back"]


class Success(AssistantSuccessPage):
    def __init__(self) -> None:
        AssistantSuccessPage.__init__(self)
        self.set_title(_("Password Changed"))
        self.set_heading(_("Password Changed"))
        self.set_text(_("Your password has successfully been changed."))

    def get_visible_buttons(self) -> list[str]:
        return ["close"]



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import datetime as dt

import pysequoia as pys
import sqlalchemy as sa
from nbxmpp import JID
from sqlalchemy import DateTime
from sqlalchemy import ForeignKey
from sqlalchemy import Index
from sqlalchemy import types
from sqlalchemy.dialects.sqlite import DATETIME
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import MappedAsDataclass
from sqlalchemy.types import TypeEngine

from gajim.common.storage.base import JIDType

DateTimeSQLite = DateTime().with_variant(DATETIME(truncate_microseconds=True), "sqlite")


class CertType(sa.types.TypeDecorator[pys.Cert]):
    impl = sa.types.BLOB
    cache_ok = True

    def process_bind_param(self, value: pys.Cert | None, dialect: Any) -> bytes | None:
        if value is None:
            return None

        return bytes(value)

    def process_result_value(
        self, value: bytes | None, dialect: Any
    ) -> pys.Cert | None:
        if value is None:
            return value
        return pys.Cert.from_bytes(value)


class Base(DeclarativeBase):
    type_annotation_map: dict[type[Any], TypeEngine[Any]] = {
        float: types.REAL(),
        str: types.TEXT(),
        int: types.INTEGER(),
    }


class Account(MappedAsDataclass, Base, kw_only=True):
    __tablename__ = "account"
    __table_args__ = (Index("idx_account", "jid", unique=True),)

    pk: Mapped[int] = mapped_column(init=False, primary_key=True)
    jid: Mapped[JID] = mapped_column(JIDType)


class Secret(MappedAsDataclass, Base, kw_only=True):
    __tablename__ = "secret"
    __table_args__ = (Index("idx_secret", "fk_account_pk", unique=True),)

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey("account.pk", ondelete="CASCADE")
    )
    key: Mapped[bytes]
    backup_password: Mapped[str | None] = mapped_column(default=None)


class Public(MappedAsDataclass, Base, kw_only=True):
    __tablename__ = "public"
    __table_args__ = (
        Index("idx_public", "fk_account_pk", "remote_jid"),
        Index(
            "idx_public_fpr", "fk_account_pk", "remote_jid", "fingerprint", unique=True
        ),
    )

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey("account.pk", ondelete="CASCADE")
    )
    remote_jid: Mapped[JID] = mapped_column(JIDType)
    key: Mapped[pys.Cert] = mapped_column(CertType)
    fingerprint: Mapped[str]
    label: Mapped[str | None] = mapped_column(default=None)
    trust: Mapped[int]
    active: Mapped[bool] = mapped_column(default=True)
    last_seen: Mapped[dt.datetime | None] = mapped_column(DateTimeSQLite, default=None)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast
from typing import get_args

import logging
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Gtk
from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common.ged import EventHelper
from gajim.common.i18n import _
from gajim.common.i18n import p_
from gajim.common.setting_values import AllSettingsT
from gajim.common.setting_values import FloatSettings

from gajim.gtk.const import Setting
from gajim.gtk.const import SettingKind
from gajim.gtk.const import SettingType
from gajim.gtk.dropdown import GajimDropDown
from gajim.gtk.filechoosers import FileChooserButton
from gajim.gtk.filechoosers import Filter
from gajim.gtk.preference.widgets import CopyButton
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.window import open_window
from gajim.gtk.window import GajimAppWindow

log = logging.getLogger("gajim.gtk.settings")


class SettingsDialog(GajimAppWindow):
    def __init__(
        self,
        parent: Gtk.Window,
        title: str,
        flags: Gtk.DialogFlags,
        settings: list[Setting],
        account: str,
    ) -> None:
        GajimAppWindow.__init__(
            self,
            name="SettingsDialog",
            title=title,
            default_width=250,
            transient_for=parent,
        )

        self.account = account
        if flags == Gtk.DialogFlags.MODAL:
            self.set_modal(True)

        elif flags == Gtk.DialogFlags.DESTROY_WITH_PARENT:
            self.set_destroy_with_parent(True)

        self._nav = Adw.NavigationView()
        self._group = GajimPreferencesGroup("main")
        page = GajimPreferencePage(title, groups=[self._group])
        self._nav.add(page)

        for setting in settings:
            self._group.add_setting(setting)

        self.set_child(self._nav)
        self.show()

    def _cleanup(self) -> None:
        del self._group

    def get_setting(self, name: str):
        return self._group.get_setting(name)


class GajimPreferencesGroup(Adw.PreferencesGroup, SignalManager, EventHelper):
    __gtype_name__ = "GajimPreferencesGroup"

    def __init__(
        self,
        key: str,
        *,
        account: str | None = None,
        jid: str | None = None,
        description: str = "",
        title: str = "",
        header_suffix: Gtk.Widget | None = None,
    ) -> None:
        EventHelper.__init__(self)
        SignalManager.__init__(self)
        Adw.PreferencesGroup.__init__(
            self, description=description, title=title, header_suffix=header_suffix
        )

        self.key = key
        self.account = account
        self.jid = jid
        self.named_settings: dict[str, GenericSetting] = {}

        self.settings_type_map: dict[SettingKind, GenericSetting] = {
            SettingKind.SWITCH: SwitchSetting,
            SettingKind.SPIN: SpinSetting,
            SettingKind.DIALOG: DialogSetting,
            SettingKind.SUBPAGE: SubPageSetting,
            SettingKind.ENTRY: EntrySetting,
            SettingKind.COLOR: ColorSetting,
            SettingKind.ACTION: ActionSetting,
            SettingKind.FILECHOOSER: FileChooserSetting,
            SettingKind.CALLBACK: CallbackSetting,
            SettingKind.DROPDOWN: DropDownSetting,
            SettingKind.GENERIC: GenericSetting,
        }

    def do_unroot(self) -> None:
        self._disconnect_all()
        self.unregister_events()
        Adw.PreferencesGroup.do_unroot(self)
        self.named_settings.clear()
        app.check_finalize(self)

    def add_setting(self, setting: Setting | Adw.ActionRow) -> None:
        if isinstance(setting, Adw.ActionRow):
            self.add(setting)
            return

        if setting.props is not None:
            row = self.settings_type_map[setting.kind](
                self.account, self.jid, *setting[1:-1], **setting.props
            )
        else:
            row = self.settings_type_map[setting.kind](
                self.account, self.jid, *setting[1:-1]
            )

        if setting.name is not None:
            self.named_settings[setting.name] = row
        self.add(row)
        row.update_activatable()

    def get_setting(self, name: str) -> GenericSetting:
        return self.named_settings[name]

    def add_copy_button(self) -> None:
        button = CopyButton()
        self._connect(button, "clicked", self._on_clipboard_button_clicked)
        self.set_header_suffix(button)

    def _on_clipboard_button_clicked(self, _widget: Gtk.Button) -> None:
        app.window.get_clipboard().set(self._get_clipboard_text())

    def _get_clipboard_text(self) -> str:
        raise NotImplementedError


class GajimPreferencePage(Adw.NavigationPage):
    key: str = ""
    icon_name: str = ""
    label: str = ""

    def __init__(
        self,
        title: str,
        groups: list[Any],
        tag_prefix: str = "",
    ) -> None:
        Adw.NavigationPage.__init__(self, tag=f"{tag_prefix}{self.key}", title=title)

        self._pref_page = Adw.PreferencesPage()
        toolbar = Adw.ToolbarView(content=self._pref_page)
        toolbar.add_top_bar(Adw.HeaderBar())
        self.set_child(toolbar)

        self._groups: list[GajimPreferencesGroup] = []

        for group in groups:
            if not isinstance(group, GajimPreferencesGroup):
                group = group()
            self._groups.append(group)
            self._pref_page.add(group)

    def do_unroot(self) -> None:
        Adw.NavigationPage.do_unroot(self)
        app.check_finalize(self)

    def add(self, group: GajimPreferencesGroup | Adw.PreferencesGroup) -> None:
        if isinstance(group, GajimPreferencesGroup):
            self._groups.append(group)
        self._pref_page.add(group)

    def get_group(self, key: str) -> GajimPreferencesGroup | None:
        for group in self._groups:
            if group.key == key:
                return group

    def set_content(self, widget: Gtk.Widget) -> None:
        toolbar = cast(Adw.ToolbarView, self.get_child())
        toolbar.set_content(widget)


class GenericSetting(Adw.ActionRow, SignalManager):
    def __init__(
        self,
        account: str,
        jid: JID,
        label: str,
        type_: SettingType,
        value: AllSettingsT | None,
        name: str | None,
        callback: Callable[..., None] | None = None,
        data: Any | None = None,
        desc: str | None = None,
        bind: str | None = None,
        inverted: bool = False,
        enabled_func: Callable[..., bool] | None = None,
        **kwargs: Any,
    ) -> None:
        Adw.ActionRow.__init__(
            self, activatable=True, title=label, subtitle=desc or "", use_markup=False
        )
        SignalManager.__init__(self)

        self.callback = callback
        self.type_ = type_
        self.value = value
        self.data = data
        self.label = label
        self.account = account
        self.jid = jid
        self.name = name
        self.bind = bind
        self.inverted = inverted
        self.enabled_func = enabled_func
        self.setting_value = self.get_value()

        self.setting_box = Gtk.Box(spacing=12, valign=Gtk.Align.CENTER)

        self._locked_icon = Gtk.Image.new_from_icon_name(
            "lucide-lock-symbolic",
        )
        self._locked_icon.set_visible(False)
        self._locked_icon.set_tooltip_text(_("Setting is locked by the system"))
        self._action_overlay = Gtk.Overlay(child=self.setting_box)
        self._action_overlay.add_overlay(self._locked_icon)

        self.add_suffix(self._action_overlay)

        self._bind_sensitive_state()

        self._add_action_button(kwargs)

        self._connect(self, "activated", self._on_activated)

    def do_unroot(self) -> None:
        Adw.ActionRow.do_unroot(self)
        self._disconnect_all()
        app.settings.disconnect_signals(self)
        del self.callback
        del self.enabled_func
        app.check_finalize(self)

    def _on_activated(self, row: Adw.ActionRow) -> None:
        return

    def _bind_sensitive_state(self) -> None:
        if self.bind is None:
            return

        bind_setting_type, setting, account, jid = self._parse_bind()

        app.settings.bind_signal(
            setting,
            self,
            "set_sensitive",
            account=account,
            jid=jid,
            inverted=self.inverted,
        )

        if bind_setting_type == SettingType.CONTACT:
            assert account is not None
            assert jid is not None
            value = app.settings.get_contact_setting(account, jid, setting)

        elif bind_setting_type == SettingType.GROUP_CHAT:
            assert account is not None
            assert jid is not None
            value = app.settings.get_group_chat_setting(account, jid, setting)

        elif bind_setting_type == SettingType.ACCOUNT_CONFIG:
            assert account is not None
            value = app.settings.get_account_setting(account, setting)

        else:
            value = app.settings.get(setting)

        assert isinstance(value, bool)
        if self.inverted:
            value = not value
        self.set_sensitive(value)

    def _parse_bind(self) -> tuple[SettingType, str, str | None, JID | None]:
        assert self.bind is not None
        if "::" not in self.bind:
            return SettingType.CONFIG, self.bind, None, None

        bind_setting_type, setting = self.bind.split("::")
        if bind_setting_type == "account":
            return SettingType.ACCOUNT_CONFIG, setting, self.account, None

        if bind_setting_type == "contact":
            return SettingType.CONTACT, setting, self.account, self.jid

        if bind_setting_type == "group_chat":
            return SettingType.GROUP_CHAT, setting, self.account, self.jid
        raise ValueError(f"Invalid bind argument: {self.bind}")

    def get_value(self) -> AllSettingsT | None:
        return self.__get_value(self.type_, self.value, self.account, self.jid)

    @staticmethod
    def __get_value(
        type_: SettingType, value: AllSettingsT | None, account: str, jid: JID
    ) -> AllSettingsT | None:
        if value is None:
            return None

        if type_ == SettingType.VALUE:
            return value

        if type_ == SettingType.CONTACT:
            return app.settings.get_contact_setting(account, jid, value)

        if type_ == SettingType.GROUP_CHAT:
            return app.settings.get_group_chat_setting(account, jid, value)

        if type_ == SettingType.CONFIG:
            return app.settings.get(value)

        if type_ == SettingType.ACCOUNT_CONFIG:
            return app.settings.get_account_setting(account, value)

        if type_ == SettingType.ACTION:
            assert isinstance(value, str)
            if value.startswith("-"):
                return account + value
            return value

        raise ValueError("Wrong SettingType?")

    def set_value(self, state: AllSettingsT) -> None:
        if self.type_ == SettingType.CONFIG:
            app.settings.set(self.value, state)

        elif self.type_ == SettingType.ACCOUNT_CONFIG:
            app.settings.set_account_setting(self.account, self.value, state)

        elif self.type_ == SettingType.CONTACT:
            app.settings.set_contact_setting(self.account, self.jid, self.value, state)

        elif self.type_ == SettingType.GROUP_CHAT:
            app.settings.set_group_chat_setting(
                self.account, self.jid, self.value, state
            )

        if self.callback is not None:
            self.callback(state, self.data)

    def update_activatable(self) -> None:
        if self.type_ == SettingType.CONFIG:
            assert isinstance(self.value, str)
            if app.settings.has_app_override(self.value):
                self.set_activatable(False)
                self.set_sensitive(False)
                self._locked_icon.set_visible(True)
                return

        if self.enabled_func is None:
            return

        enabled_func_value = self.enabled_func()
        self.set_activatable(enabled_func_value)
        self.set_sensitive(enabled_func_value)

    def _add_action_button(
        self, kwargs: dict[str, str | Callable[..., None] | None]
    ) -> None:
        icon_name = kwargs.get("button-icon-name")
        button_text = kwargs.get("button-text")
        tooltip_text = cast(str, kwargs.get("button-tooltip") or "")
        style = kwargs.get("button-style")

        if icon_name is not None:
            assert isinstance(icon_name, str)
            button = Gtk.Button.new_from_icon_name(icon_name)

        elif button_text is not None:
            assert isinstance(button_text, str)
            button = Gtk.Button(label=button_text)

        else:
            return

        if style is not None:
            assert isinstance(style, str)
            for css_class in style.split(" "):
                button.add_css_class(css_class)

        callback = kwargs["button-callback"]
        assert isinstance(callback, Callable)
        self._connect(button, "clicked", callback)
        button.set_tooltip_text(tooltip_text)

        sensitive = bool(kwargs.get("button-sensitive", True))
        button.set_sensitive(sensitive)

        self.setting_box.append(button)


class SwitchSetting(GenericSetting):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        GenericSetting.__init__(self, *args, **kwargs)

        self.switch = Gtk.Switch()
        if self.type_ == SettingType.ACTION:
            self.switch.set_action_name(f"app.{self.setting_value}")
            assert isinstance(self.setting_value, str)
            state = app.app.get_action_state(self.setting_value)
            assert state is not None
            self.switch.set_active(state.get_boolean())
        else:
            assert isinstance(self.setting_value, bool)
            self.switch.set_active(self.setting_value)
        self._connect(self.switch, "notify::active", self._on_active_notify)
        self.switch.set_hexpand(True)
        self.switch.set_halign(Gtk.Align.END)
        self.switch.set_valign(Gtk.Align.CENTER)

        self._switch_state_label = Gtk.Label()
        self._switch_state_label.set_xalign(1)
        self._switch_state_label.set_valign(Gtk.Align.CENTER)
        assert isinstance(self.setting_value, bool)
        self._set_label(self.setting_value)

        box = Gtk.Box(spacing=12)
        box.set_halign(Gtk.Align.END)
        box.append(self._switch_state_label)
        box.append(self.switch)
        self.setting_box.prepend(box)

        self.set_activatable_widget(self.switch)

    def _on_active_notify(self, switch: Gtk.Switch, *args: Any) -> None:
        value = switch.get_active()
        self.set_value(value)
        self._set_label(value)

    def _set_label(self, active: bool) -> None:
        text = p_("Switch", "On") if active else p_("Switch", "Off")
        self._switch_state_label.set_text(text)


class EntrySetting(GenericSetting):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        GenericSetting.__init__(self, *args, **kwargs)

        self.entry = Gtk.Entry()
        self.entry.set_text(str(self.setting_value))
        self._changed_handler_id = self._connect(
            self.entry, "changed", self._on_text_change
        )
        self.entry.set_valign(Gtk.Align.CENTER)
        self.entry.set_alignment(1)

        self.setting_box.prepend(self.entry)

        assert isinstance(self.value, str)
        app.settings.connect_signal(
            self.value, self._on_setting_changed, account=self.account, jid=self.jid
        )

        self.set_activatable_widget(self.entry)

    def _on_setting_changed(self, value: str, *args: Any) -> None:
        if self.entry.get_text() == value:
            return

        with self.entry.handler_block(self._changed_handler_id):
            self.entry.set_text(value)

    def _on_text_change(self, *args: Any) -> None:
        text = self.entry.get_text()
        self.set_value(text)


class ColorSetting(GenericSetting):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        GenericSetting.__init__(self, *args, **kwargs)

        rgba = Gdk.RGBA()
        assert isinstance(self.setting_value, str)
        rgba.parse(self.setting_value)
        color_dialog = Gtk.ColorDialog()
        self.color_button = Gtk.ColorDialogButton(dialog=color_dialog)
        self.color_button.set_rgba(rgba)
        self._connect(self.color_button, "notify::rgba", self._on_color_set)
        self.color_button.set_valign(Gtk.Align.CENTER)
        self.color_button.set_halign(Gtk.Align.END)
        self.color_button.set_hexpand(True)

        self.setting_box.prepend(self.color_button)

        assert isinstance(self.value, str)
        app.settings.connect_signal(
            self.value, self._on_setting_changed, account=self.account, jid=self.jid
        )

        self.set_activatable_widget(self.color_button)

    def _on_setting_changed(self, value: str, *args: Any) -> None:
        rgba = Gdk.RGBA()
        rgba.parse(value)
        self.color_button.set_rgba(rgba)

    def _on_color_set(self, color_button: Gtk.ColorDialogButton, *args: Any) -> None:
        rgba = color_button.get_rgba()
        self.set_value(rgba.to_string())
        app.css_config.refresh()


@dataclass
class SpinRange:
    lower: float
    upper: float
    step: float
    digits: int = 0


class SpinSetting(GenericSetting):
    def __init__(
        self, *args: Any, range_: tuple[float, float, float] | SpinRange, **kwargs: Any
    ) -> None:
        GenericSetting.__init__(self, *args, **kwargs)

        if not isinstance(range_, SpinRange):
            range_ = SpinRange(*range_)

        adjustment = Gtk.Adjustment(
            value=0,
            lower=range_.lower,
            upper=range_.upper,
            step_increment=range_.step,
            page_increment=10,
            page_size=0,
        )

        self.spin = Gtk.SpinButton(
            adjustment=adjustment,
            halign=Gtk.Align.END,
            hexpand=True,
            numeric=True,
            update_policy=Gtk.SpinButtonUpdatePolicy.IF_VALID,
            valign=Gtk.Align.CENTER,
            width_chars=5,
            digits=range_.digits,
        )

        assert isinstance(self.setting_value, int | float | str)
        self.spin.set_value(float(self.setting_value))

        self._connect(self.spin, "notify::value", self._on_value_change)

        self.setting_box.prepend(self.spin)

        assert isinstance(self.value, str)
        app.settings.connect_signal(
            self.value, self._on_setting_changed, account=self.account, jid=self.jid
        )

        self.set_activatable_widget(self.spin)

    def _on_setting_changed(self, value: float, *args: Any) -> None:
        self.spin.set_value(value)

    def _on_value_change(self, spin: Gtk.SpinButton, *args: Any) -> None:
        if self.value in list(get_args(FloatSettings)):
            value = spin.get_value()
        else:
            value = spin.get_value_as_int()
        self.set_value(value)


class FileChooserSetting(GenericSetting):
    def __init__(self, *args: Any, filefilters: list[Filter], **kwargs: Any) -> None:
        GenericSetting.__init__(self, *args, **kwargs)
        button = FileChooserButton(filters=filefilters, label=self.label)
        button.set_halign(Gtk.Align.END)

        if self.setting_value:
            assert isinstance(self.setting_value, str)
            button.set_path(Path(self.setting_value))

        self._connect(button, "path-picked", self._on_select)

        clear_button = Gtk.Button(
            icon_name="lucide-brush-cleaning-symbolic", tooltip_text=_("Clear File")
        )
        self._connect(clear_button, "clicked", lambda *args: button.reset())  # type: ignore
        self.setting_box.prepend(clear_button)
        self.setting_box.prepend(button)

        self.set_activatable_widget(button)

    def _on_select(
        self, _file_chooser_button: FileChooserButton, file_paths: list[Path]
    ) -> None:
        if not file_paths:
            return

        self.set_value(str(file_paths[0]))


class CallbackSetting(GenericSetting):
    def __init__(self, *args: Any, callback: Callable[..., Any], **kwargs: Any) -> None:
        GenericSetting.__init__(self, *args, **kwargs)
        self.callback = callback

    def _on_activated(self, row: Adw.ActionRow) -> None:
        self.callback()


class ActionSetting(GenericSetting):
    def __init__(self, *args: Any, variant: GLib.Variant, **kwargs: Any) -> None:
        GenericSetting.__init__(self, *args, **kwargs)
        assert isinstance(self.value, str)
        if self.value.startswith("app."):
            self.action = app.app.lookup_action(self.value[4:])
        else:
            self.action = app.window.lookup_action(self.value)

        if self.action is None:
            log.error("Action not found: %s", self.value)
            return

        self.variant = variant
        self.on_enable()

        self._connect(self.action, "notify::enabled", self.on_enable)

    def on_enable(self, *args: Any) -> None:
        if self.action is not None:
            self.set_sensitive(self.action.get_enabled())

    def _on_activated(self, row: Adw.ActionRow) -> None:
        if self.action is not None:
            self.action.activate(self.variant)


class DropDownSetting(GenericSetting):
    def __init__(
        self, *args: Any, data: list[str] | dict[str, Any], **kwargs: Any
    ) -> None:
        GenericSetting.__init__(self, *args, **kwargs)

        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=12)
        box.set_halign(Gtk.Align.END)
        box.set_hexpand(True)

        self._dropdown: GajimDropDown[Any] = GajimDropDown(data=data, fixed_width=15)
        self._dropdown.select_key(self.setting_value)
        self._dropdown.connect("notify::selected", self._on_selected)
        box.append(self._dropdown)

        self.setting_box.prepend(box)

        self.set_activatable_widget(self._dropdown)

    def do_unroot(self) -> None:
        self._dropdown.disconnect_by_func(self._on_selected)
        GenericSetting.do_unroot(self)
        self._dropdown.run_destroy()
        del self._dropdown

    def _on_selected(self, dropdown: GajimDropDown[Any], *args: Any) -> None:
        item = dropdown.get_selected_item()
        assert item is not None
        self.set_value(item.key)

    def update_entries(self, data: list[str] | dict[str, str]) -> None:
        self._dropdown.disconnect_by_func(self._on_selected)
        self._dropdown.set_data(data)
        self._dropdown.connect("notify::selected", self._on_selected)

    def select_key(self, key: Any) -> None:
        self._dropdown.disconnect_by_func(self._on_selected)
        self._dropdown.select_key(key)
        self._dropdown.connect("notify::selected", self._on_selected)


class DialogSetting(GenericSetting):
    def __init__(self, *args: Any, dialog: str, **kwargs: Any) -> None:
        GenericSetting.__init__(self, *args, **kwargs)

        self._dialog = dialog

        image = Gtk.Image.new_from_icon_name("lucide-chevron-right-symbolic")
        self.add_suffix(image)

    def _on_activated(self, row: Adw.ActionRow) -> None:
        open_window(self._dialog, account=self.account)  # type: ignore


class SubPageSetting(GenericSetting):
    def __init__(self, *args: Any, subpage: str, **kwargs: Any) -> None:
        GenericSetting.__init__(self, *args, **kwargs)

        self._subpage = subpage

        self._label = Gtk.Label()
        self.add_suffix(self._label)

        image = Gtk.Image.new_from_icon_name("lucide-chevron-right-symbolic")
        self.add_suffix(image)

        if self.value is not None:
            self.set_label(self.setting_value)
            app.settings.bind_signal(
                self.value,
                self,
                "set_label",
                account=self.account,
                jid=self.jid,
                inverted=self.inverted,
            )

    def set_label(self, enabled: bool) -> None:
        text = p_("Switch", "On") if enabled else p_("Switch", "Off")
        self._label.set_text(text)

    def _on_activated(self, row: Adw.ActionRow) -> None:
        nav = row.get_ancestor(Adw.NavigationView)
        if nav is None:
            return

        nav.activate_action("navigation.push", GLib.Variant("s", self._subpage))
